#ifndef __FSDB_DATE_TIME_H__
#define __FSDB_DATE_TIME_H__

/*
This code is taken from the IBPP library
www.ibpp.org
*/

#include <stdexcept>

namespace FSDB {

	// substitute for the same IBPP class
	class LogicExceptionImpl : public std::out_of_range {
	public:
		LogicExceptionImpl(const std::string& s ) : std::out_of_range( s ) {}
	};

	/* Classes Date, Time, Timestamp and DBKey are 'helper' classes.  They help
	* in retrieving or setting some special SQL types. Dates, times and dbkeys
	* are often read and written as strings in SQL scripts. When programming
	* with IBPP, we handle those data with these specific classes, which
	* enhance their usefullness and free us of format problems (M/D/Y, D/M/Y,
	* Y-M-D ?, and so on...). */

	/* Class Date represent purely a Date (no time part specified). It is
	* usefull in interactions with the SQL DATE type of Interbase.  You can add
	* or substract a number from a Date, that will modify it to represent the
	* correct date, X days later or sooner. All the Y2K details taken into
	* account.
	* The full range goes from integer values IBPP::MinDate to IBPP::MaxDate
	* which means from 01 Jan 0001 to 31 Dec 9999. ( Which is inherently
	* incorrect as this assumes Gregorian calendar. ) */

	//	Dates range checking
	const int MinDate = -693594;	//  1 JAN 0001
	const int MaxDate = 2958464;	// 31 DEC 9999

	class Timestamp;	// Cross-reference between Timestamp, Date and Time

	class Date
	{
	protected:
		int mDate;	// The date : 1 == 1 Jan 1900

	public:
		void Clear()	{ mDate = MinDate - 1; };
		void Today();
		void SetDate(int year, int month, int day);
		void SetDate(int dt);
		void GetDate(int& year, int& month, int& day) const;
		int GetDate() const	{ return mDate; }
		int Year() const;
		int Month() const;
		int Day() const;
		void Add(int days);
		void StartOfMonth();
		void EndOfMonth();

		Date()			{ Clear(); };
		Date(int dt)	{ SetDate(dt); }
		Date(int year, int month, int day);
		Date(const Date&);							// Copy Constructor
		Date& operator=(const Timestamp&);			// Timestamp Assignment operator
		Date& operator=(const Date&);				// Date Assignment operator

		bool operator==(const Date& rv)	const { return mDate == rv.GetDate(); }
		bool operator!=(const Date& rv)	const { return mDate != rv.GetDate(); }
		bool operator<(const Date& rv) const { return mDate < rv.GetDate(); }
		bool operator>(const Date& rv) const { return mDate > rv.GetDate(); }

		virtual ~Date() { };
	};

	/* Class Time represent purely a Time. It is usefull in interactions
	* with the SQL TIME type of Interbase. */

	class Time
	{
	protected:
		int mTime;	// The time, in ten-thousandths of seconds since midnight

	public:
		void Clear()	{ mTime = 0; }
		void Now();
		void SetTime(int hour, int minute, int second, int tenthousandths = 0);
		void SetTime(int tm);
		void GetTime(int& hour, int& minute, int& second) const;
		void GetTime(int& hour, int& minute, int& second, int& tenthousandths) const;
		int GetTime() const	{ return mTime; }
		int Hours() const;
		int Minutes() const;
		int Seconds() const;
		int SubSeconds() const;		// Actually tenthousandths of seconds
		Time()			{ Clear(); }
		Time(int tm)	{ SetTime(tm); }
		Time(int hour, int minute, int second, int tenthousandths = 0);
		Time(const Time&);							// Copy Constructor
		Time& operator=(const Timestamp&);			// Timestamp Assignment operator
		Time& operator=(const Time&);				// Time Assignment operator

		bool operator==(const Time& rv)	const { return mTime == rv.GetTime(); }
		bool operator!=(const Time& rv)	const { return mTime != rv.GetTime(); }
		bool operator<(const Time& rv) const { return mTime < rv.GetTime(); }
		bool operator>(const Time& rv) const { return mTime > rv.GetTime(); }

		virtual ~Time() { };
	};

	/* Class Timestamp represent a date AND a time. It is usefull in
	* interactions with the SQL TIMESTAMP type of Interbase. This class
	* inherits from Date and Time and completely inline implements its small
	* specific details. */

	class Timestamp : public Date, public Time
	{
	public:
		void Clear()	{ Date::Clear(); Time::Clear(); }
		void Today()	{ Date::Today(); Time::Clear(); }
		void Now()		{ Date::Today(); Time::Now(); }

		Timestamp()		{ Clear(); }

		Timestamp(int y, int m, int d)
		{ Date::SetDate(y, m, d); Time::Clear(); }

		Timestamp(int y, int mo, int d, int h, int mi, int s, int t = 0)
		{ Date::SetDate(y, mo, d); Time::SetTime(h, mi, s, t); }

		Timestamp(const Timestamp& rv)
			: Date(rv.mDate), Time(rv.mTime) {}	// Copy Constructor

			Timestamp(const Date& rv)
			{ mDate = rv.GetDate(); mTime = 0; }

			Timestamp(const Time& rv)
			{ mDate = 0; mTime = rv.GetTime(); }

			Timestamp& operator=(const Timestamp& rv)	// Timestamp Assignment operator
			{ mDate = rv.mDate; mTime = rv.mTime; return *this; }

			Timestamp& operator=(const Date& rv)		// Date Assignment operator
			{ mDate = rv.GetDate(); return *this; }

			Timestamp& operator=(const Time& rv)		// Time Assignment operator
			{ mTime = rv.GetTime(); return *this; }

			bool operator==(const Timestamp& rv) const
			{ return (mDate == rv.GetDate()) && (mTime == rv.GetTime()); }

			bool operator!=(const Timestamp& rv) const
			{ return (mDate != rv.GetDate()) || (mTime != rv.GetTime()); }

			bool operator<(const Timestamp& rv) const
			{ return (mDate < rv.GetDate()) ||
			(mDate == rv.GetDate() && mTime < rv.GetTime()); }

			bool operator>(const Timestamp& rv) const
			{ return (mDate > rv.GetDate()) ||
			(mDate == rv.GetDate() && mTime > rv.GetTime()); }

			~Timestamp() { }
	};

	/* Finally, here are some date and time conversion routines used by IBPP and
	 * that may be helpful at the application level. They do not depend on
	 * anything related to Firebird/Interbase. Just a bonus. dtoi and itod
	 * return false on invalid parameters or out of range conversions. */

	bool dtoi(int date, int* py, int* pm, int* pd);
	bool itod(int* pdate, int year, int month, int day);
	void ttoi(int itime, int* phour, int* pminute, int* psecond, int* ptt);
	void itot(int* ptime, int hour, int minute, int second = 0, int tenthousandths = 0);
}


#endif


