/*
This file is part of FSDB: a higher level library to access a Firebird database using IBPP.

Copyright Fulvio Senore.

License:

Permission is hereby granted, free of charge, to any person or organization ("You") obtaining a copy of this software and 
associated documentation files covered by this license (the "Software") to use the Software as part of another work; 
to modify it for that purpose; to publish or distribute it, modified or not, for that same purpose; 
to permit persons to whom the other work using the Software is furnished to do so; subject to the following conditions: 
the above copyright notice and this complete and unmodified permission notice shall be included in all copies or substantial 
portions of the Software; You will not misrepresent modified versions of the Software as being the original.

The Software is provided "as is", without warranty of any kind, express or implied, including but not limited 
to the warranties of merchantability, fitness for a particular purpose and noninfringement. 
In no event shall the authors or copyright holders be liable for any claim, damages or other liability, 
whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software 
or the use of other dealings in the Software.
*/


#ifndef __FSDB_INTERFACE_BASE_H__
#define __FSDB_INTERFACE_BASE_H__

#include "_fsdb.h"
#include "variant.h"
#include "parameters.h"

#include <string>

namespace FSDB {

	/** Defines the interface to a database.
	Abstract class that defines the interface to a particular database.
	One object of a class derived from this one is contained in each recordset object
	and it handles communication with the database.
	*/
	class DbInterfaceBase {

	public:

		virtual ~DbInterfaceBase() {};

		/// open a recordset from a SQL string.
		/// "params" may be null for a query without parameters
		virtual void OpenRecordset( RecordsetType rt, const std::string &sql, Parameters params, const std::string &tableName = "", const std::string &primaryKeyName = "" ) = 0;

		virtual void Close() = 0;
		/// move to the next row
		virtual void MoveNext() = 0;

		/// returns a variant containing the column value in the current row
		virtual Variant& GetVariantByName( const std::string& columnName ) = 0;

		/// returns the zero-based position of the column in the row
		virtual int GetColumnPosition( const std::string& columnName ) = 0;

		/// returns true if we moved past the last row (EOF)
		virtual bool IsEOF() = 0;
		/// returns true if we moved before the first row (BOF)
		virtual bool IsBOF() = 0;
		/// returns true if the recordset is currently opened
		virtual bool IsRecordsetOpened() = 0;

		/// returns the number of columns in the currently opened recordset
		virtual int GetNColumns() = 0;

		/// reads the current row in the rowbuffer.
		/// the row variable must be allocated and deallocated by the caller
		virtual void ReadCurrentRow( Variant *row ) = 0;

		/// return a SQL string to insert a new row.
		/// return an empty string if there is nothing to do (no changed columns)
		virtual std::string CreateSQLStringForInsert( Variant *row ) = 0;

		/// return a string to apply the update of the changed columns.
		/// return an empty string if there is nothing to do (no changed columns)
		virtual std::string CreateSQLStringForUpdate( Variant *row ) = 0;

		/// implements the RecordsetBase::Update() method for a previous AddNew()
		virtual void UpdateAddNew( Variant *row ) = 0;
		/// implements the RecordsetBase::Update() method for a previous Edit()
		virtual void UpdateEdit( Variant *row ) = 0;

        /// delete the current row
        virtual void Delete( Variant *row ) = 0;

	};


}



#endif

