/*
This file is part of FSDB: a higher level library to access a Firebird database using IBPP.

Copyright Fulvio Senore.

License:

Permission is hereby granted, free of charge, to any person or organization ("You") obtaining a copy of this software and 
associated documentation files covered by this license (the "Software") to use the Software as part of another work; 
to modify it for that purpose; to publish or distribute it, modified or not, for that same purpose; 
to permit persons to whom the other work using the Software is furnished to do so; subject to the following conditions: 
the above copyright notice and this complete and unmodified permission notice shall be included in all copies or substantial 
portions of the Software; You will not misrepresent modified versions of the Software as being the original.

The Software is provided "as is", without warranty of any kind, express or implied, including but not limited 
to the warranties of merchantability, fitness for a particular purpose and noninfringement. 
In no event shall the authors or copyright holders be liable for any claim, damages or other liability, 
whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software 
or the use of other dealings in the Software.
*/


#include "db_interface_firebird.h"
#include "data_error.h"

#include <stdexcept>
#include <assert.h>

#define BLOB_BUFSIZE 8192



FSDB::DbInterfaceFirebird::~DbInterfaceFirebird() {
}


void FSDB::DbInterfaceFirebird::OpenRecordset( RecordsetType rt, const std::string &sql, Parameters params, const std::string &tableName, const std::string &primaryKeyName ) {

	if( isOpened ) throw std::runtime_error( "Recordset already opened" );

	switch( rt ) {
		case rtReadOnlyForwardOnly:
		case rtReadWriteForwardOnly:
		case rtReadWrite:
			// if the user has already opened an explicit transaction we use it
			if( db->IsTransactionPending() ) {
				tr = db->GetIBPPTransaction();
			}
			else {
				// we open a read only transaction also for rtReadWrite because writing is
				// implemented in another (short-lived) transaction
				tr = IBPP::TransactionFactory(db->GetIBPPDatabase(), IBPP::amRead, IBPP::ilReadCommitted );
				tr->Start();
			}
			break;
		case rtReadOnlyRepeatableForwardOnly:
			if( db->IsTransactionPending() ) {
				throw std::runtime_error( "Unable to open a repeteable forward only recordset inside an expliciti transaction" );
			}
			else {
				tr = IBPP::TransactionFactory(db->GetIBPPDatabase(), IBPP::amRead, IBPP::ilConcurrency );
				tr->Start();
			}
			break;
		default:
			throw std::invalid_argument( "Unhandled recordset type" );
			break;
	}

	st = IBPP::StatementFactory( db->GetIBPPDatabase(), tr );
	try {
		if( params == NULL ) {
			st->Execute( sql );
		}
		else {
			st->Prepare( sql );
			db->SetStatementParameters( st, params );
			st->Execute();
		}
	}
	catch( IBPP::SQLException& e ) {
		throw CDataErrorException::IBPPSql2CDataError( e );
	}
	catch( IBPP::LogicException& e ) {
		throw CDataErrorException::IBPPLogic2CDataError( e );
	}
	catch(...) {
		throw;
	}

	m_eof = !st->Fetch();
	m_bof = m_eof;

	if( rt == rtReadWrite || rt == rtReadWriteForwardOnly ) {
		if( tableName.empty() || primaryKeyName.empty() )
			throw std::invalid_argument( "You must supply tableName and primaryKeyName for a read-write recordset" );

		this->tableName = tableName;
		this->primaryKeyName = primaryKeyName;
		primaryKeyColumn = st->ColumnNum( primaryKeyName );
	}

	isOpened = true;
}

int FSDB::DbInterfaceFirebird::GetNColumns() {
	if( !isOpened ) throw std::runtime_error( "Recordset not opened" );
	return st->Columns();
}

void FSDB::DbInterfaceFirebird::Close() {
	if( !isOpened ) throw std::runtime_error( "Recordset already closed" );
	st->Close();
	if( !db->IsTransactionPending() ) tr->Commit();
	isOpened = false;
}

void FSDB::DbInterfaceFirebird::MoveNext() {
	m_eof = !st->Fetch();
}

void FSDB::DbInterfaceFirebird::ReadBlob( Variant &var, int colIndex ) {
	char *buf = new char[BLOB_BUFSIZE];

    // do not store the original value when reading from the database
    bool curHandleOriginalValue = var.GetHandleOriginalValue();
    var.SetHandleOriginalValue( false );

    IBPP::Blob b = IBPP::BlobFactory( db->GetIBPPDatabase(), tr );
	st->Get( colIndex, b );
	b->Open();
	int size, largest, segments;
	b->Info( &size, &largest, &segments );
	char *blobData = new char[size];

	int read = 0;
	int nChars = 1;	// dummy value
	char *cp = blobData;
	while( nChars > 0 ) {
		nChars = b->Read( cp, BLOB_BUFSIZE );
		cp += nChars;
		read += nChars;
	}

	b->Close();
	var.SetAsBlobNoMemoryAllocation( blobData, size );

    var.SetHandleOriginalValue( curHandleOriginalValue );

	delete buf;
}

void FSDB::DbInterfaceFirebird::ReadColumn( Variant &var, int colIndex )
{
	IBPP::SDT IBPP_Type = st->ColumnType( colIndex );

    // do not store the original value when reading from the database
    bool curHandleOriginalValue = var.GetHandleOriginalValue();
    var.SetHandleOriginalValue( false );

	if( st->IsNull(colIndex) ) {
		var.SetNull();
        var.SetHandleOriginalValue( curHandleOriginalValue );
		return;
	}

	switch( IBPP_Type ) {
		case IBPP::sdFloat:
			{
				float tmp;
				st->Get( colIndex, tmp );
				var = tmp;
			}
			break;
		case IBPP::sdDouble:
			{
				double tmp;
				st->Get( colIndex, tmp );
				var = tmp;
			}
			break;
			case IBPP::sdString:
			{
				std::string tmp;
				st->Get( colIndex, tmp );
				var = tmp;
			}
			break;
		case IBPP::sdSmallint:
			{
				int16_t tmp;
				st->Get( colIndex, tmp );
				var = tmp;
			}
			break;
		case IBPP::sdInteger:
			{
				int32_t tmp;
				st->Get( colIndex, tmp );
				var = tmp;
			}
			break;
		case IBPP::sdLargeint:
			{
				int64_t tmp;
				st->Get( colIndex, tmp );
				var = tmp;
			}
			break;
		case IBPP::sdBlob:
			ReadBlob( var, colIndex );
			break;
		case IBPP::sdTimestamp:
			{
				IBPP::Timestamp ts;
				st->Get( colIndex, ts );
				FSDB::Timestamp tmp( ts.Year(), ts.Month(), ts.Day(), ts.Hours(), ts.Minutes(), ts.Seconds(), ts.SubSeconds() );
				var = tmp;
			}
			break;
	}

    var.SetHandleOriginalValue( curHandleOriginalValue );
}

FSDB::Variant& FSDB::DbInterfaceFirebird::GetVariantByName( const std::string& columnName ) {
    int colIndex;
    try {
        colIndex = st->ColumnNum( columnName );
    }
	catch( IBPP::LogicException& e ) {
        std::string moreInfo = "column name: " + columnName;
		throw CDataErrorException::IBPPLogic2CDataError( e, moreInfo );
	}
	catch(...) {
		throw;
	}
	ReadColumn( tmpVariant, colIndex );
	return tmpVariant;
}

void FSDB::DbInterfaceFirebird::ReadCurrentRow( Variant *row ) {

	for( int k = 0; k < GetNColumns(); k ++ ) {
		ReadColumn( row[k], k + 1 );	// IBPP columns start from 1, not 0
		row[k].ClearChanged();
	}
}

int FSDB::DbInterfaceFirebird::GetColumnPosition( const std::string& columnName )
{
    int colIndex;
    try {
        colIndex = st->ColumnNum( columnName );
    }
	catch( IBPP::LogicException& e ) {
        std::string moreInfo = "column name: " + columnName;
		throw CDataErrorException::IBPPLogic2CDataError( e, moreInfo );
	}
	catch(...) {
		throw;
	}

    return colIndex - 1;
}

std::string FSDB::DbInterfaceFirebird::CreateSQLStringForUpdate( Variant *row ) {
	std::string sql;
	bool foundSomething;
	bool firstTime;

	foundSomething = false;
	firstTime = true;
	sql = "UPDATE " + tableName + " SET ";

	for( int k = 0; k < GetNColumns(); k++ ) {
		if( row[k].IsChanged() ) {
			if( !firstTime ) sql += ", ";
			sql = sql + st->ColumnName(k + 1) + " = " + db->ConvertToSQLString( row[k] );
			foundSomething = true;
			firstTime = false;
		}
	}
	
	Variant primaryKeyValue = row[primaryKeyColumn - 1];
    if( primaryKeyValue.IsChanged() ) {
        // they changed the primary key, so we need to get the original value to locate the row that will be updated
        primaryKeyValue.RestoreOriginalValue();
    }
    sql += " WHERE " + primaryKeyName + " = " + db->ConvertToSQLString( primaryKeyValue );

	if( foundSomething )
		return sql;
	else
		return "";
}

std::string FSDB::DbInterfaceFirebird::CreateSQLStringForInsert( Variant *row ) {
	std::string sql, columns, values;
	bool firstTime;

	sql = "";

	firstTime = true;
	columns = "";
	values = "";
	for( int k = 0; k < GetNColumns(); k++ ) {
		if( !row[k].IsNull() ) {
		    if( !firstTime ) {
			    columns += ", ";
			    values += ", ";
		    }
		    columns += st->ColumnName(k + 1);
		    values += db->ConvertToSQLString( row[k] );
		    firstTime = false;
        }
	}

	if( columns != "" ) {
		sql = "INSERT INTO " + tableName + " (" + columns + ") VALUES (" + values + ")";
	}

	return sql;
}

void FSDB::DbInterfaceFirebird::UpdateAddNew( Variant *row )
{
	std::string sql = CreateSQLStringForInsert( row );

	if( sql == "" )
		throw std::runtime_error( "Called Update() with nothing to store in the database" );

	std::vector<BlobData> bdv;
	LoadBlobData( row, bdv );

	// executes the sql string
    ExecuteSQLString( sql, bdv );

    // clear the changed flag for the written data
	for( int k = 0; k < GetNColumns(); k++ ) {
        row[k].ClearChanged();
    }
}

void FSDB::DbInterfaceFirebird::UpdateEdit( Variant *row ) {

	std::string sql = CreateSQLStringForUpdate( row );

	if( sql == "" )
		throw std::runtime_error( "Called Update() with nothing to store in the database" );

	std::vector<BlobData> bdv;
	LoadBlobData( row, bdv );

	// executes the sql string
    try {
        ExecuteSQLString( sql, bdv );
    }
    catch(...) {
        // restore the original values, since we could not write the new data
	    for( int k = 0; k < GetNColumns(); k++ ) {
            row[k].RestoreOriginalValue();
        }
        throw;
    }

    // clear the changed flag for the written data
	for( int k = 0; k < GetNColumns(); k++ ) {
        row[k].ClearChanged();
    }
}

void FSDB::DbInterfaceFirebird::LoadBlobData( Variant *row, std::vector<BlobData> &bdv ) {

	for( int k = 0; k < GetNColumns(); k++ ) {
		if( row[k].GetCurrentType() == FSDB::Variant::ctBlob ) {
			BlobData *bd = new BlobData();
			row[k].GetAsBlobConstPointer( bd->cp, bd->size );
			bdv.push_back( *bd );
			delete bd;
		}
	}
}

IBPP::Blob FSDB::DbInterfaceFirebird::WriteBlob( BlobData bd, IBPP::Transaction m_tr ) {
	char *buf = new char[BLOB_BUFSIZE];

	IBPP::Blob b = IBPP::BlobFactory( db->GetIBPPDatabase(), m_tr );
	b->Create();
	int written = 0;
	const char *cp = (const char*) bd.cp;
	while( written < bd.size ) {
		int nChars = BLOB_BUFSIZE;
		if( written + nChars > bd.size )
			nChars = bd.size - written;
		b->Write( cp, nChars );
		written += nChars;
		cp += nChars;
	}

	delete buf;
	b->Close();
	return b;
}

void FSDB::DbInterfaceFirebird::ExecuteSQLString( const std::string &sql, const std::vector<BlobData> &bdv ) {
	IBPP::Transaction m_tr;
	IBPP::Statement m_st;

	assert( sql != "" );

	if( !db->IsTransactionPending() ) {
		m_tr = TransactionFactory( db->GetIBPPDatabase(), IBPP::amWrite );
		m_tr->Start();
	}
	else {
		m_tr = db->GetIBPPTransaction();
	}

	m_st = StatementFactory( db->GetIBPPDatabase(), m_tr );
	try {
		if( bdv.size() == 0 ) {
			m_st->Execute( sql );
		}
		else {
			m_st->Prepare( sql );
			for( unsigned int k = 0; k < bdv.size(); k++ ) {
				IBPP::Blob b = WriteBlob( bdv[k], m_tr );
				m_st->Set( k + 1, b );
			}
			m_st->Execute();
		}
	}
	catch( IBPP::SQLException& e ) {
	    if( !db->IsTransactionPending() ) {
            m_tr->Rollback();
	    }
		throw CDataErrorException::IBPPSql2CDataError( e );
	}
	catch( IBPP::LogicException& e ) {
	    if( !db->IsTransactionPending() ) {
            m_tr->Rollback();
	    }
		throw CDataErrorException::IBPPLogic2CDataError( e );
	}
	catch(...) {
		throw;
	}

	if( !db->IsTransactionPending() ) {
		m_tr->Commit();
	}

}

void FSDB::DbInterfaceFirebird::Delete( Variant *row )
{
    std::string sql;

	Variant primaryKeyValue = row[primaryKeyColumn - 1];
    sql = "DELETE FROM " + tableName + " WHERE " + primaryKeyName + " = " + db->ConvertToSQLString( primaryKeyValue );
	std::vector<BlobData> bdv;
    ExecuteSQLString( sql, bdv );
}

