/*
This file is part of FSDB: a higher level library to access a Firebird database using IBPP.

Copyright Fulvio Senore.

License:

Permission is hereby granted, free of charge, to any person or organization ("You") obtaining a copy of this software and 
associated documentation files covered by this license (the "Software") to use the Software as part of another work; 
to modify it for that purpose; to publish or distribute it, modified or not, for that same purpose; 
to permit persons to whom the other work using the Software is furnished to do so; subject to the following conditions: 
the above copyright notice and this complete and unmodified permission notice shall be included in all copies or substantial 
portions of the Software; You will not misrepresent modified versions of the Software as being the original.

The Software is provided "as is", without warranty of any kind, express or implied, including but not limited 
to the warranties of merchantability, fitness for a particular purpose and noninfringement. 
In no event shall the authors or copyright holders be liable for any claim, damages or other liability, 
whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software 
or the use of other dealings in the Software.
*/

/*
Implementation for a Firebird database
*/

#ifndef __FSDB_INTERFACE_FIREBIRD_H__
#define __FSDB_INTERFACE_FIREBIRD_H__

#include "db_interface_base.h"
#include "database_firebird.h"
#include "variant.h"

#include "../ibpp/ibpp.h"

#include <vector>

namespace FSDB {

	/// Database interface for a Firebird database
	class DbInterfaceFirebird : public DbInterfaceBase {

	public:

		DbInterfaceFirebird( FSDB::DatabaseFirebird* db ) : db(db), m_eof(true), m_bof(true), isOpened(false) {}
		~DbInterfaceFirebird();

		void OpenRecordset( RecordsetType rt, const std::string &sql, Parameters params, const std::string &tableName = "", const std::string &primaryKeyName = "" );

		void Close();
		bool IsEOF() { return m_eof; }
		bool IsBOF() { return m_bof; }
		void MoveNext();
		bool IsRecordsetOpened() { return isOpened; }

		Variant& GetVariantByName( const std::string& columnName );
		int GetColumnPosition( const std::string& columnName );

		int GetNColumns();

		void ReadCurrentRow( Variant *row );

		void UpdateAddNew( Variant *row );
		void UpdateEdit( Variant *row );

        void Delete( Variant *row );

    protected:

		/// Contains the data about a blob: a pointer and its size.
		struct BlobData {
			const void *cp;
			int size;
		};

		IBPP::Transaction tr;
		IBPP::Statement st;
		FSDB::DatabaseFirebird* db;

		/// end of file flag
		bool m_eof;
		// BOF flag
		bool m_bof;

		Variant tmpVariant;

		/// name of the table that will be updated: used only for updatable recordsets
		std::string tableName;
		/// name of the primary key: used only for updatable recordsets
		std::string primaryKeyName;
		/// column number of the primary key
		int primaryKeyColumn;

		/// true if the recordset is currently opened
		bool isOpened;

		/// read a column from the current database row to the passed Variant instance
		void ReadColumn( Variant &var, int colIndex );

		/// read a blob in the Variant
		void ReadBlob( Variant &var, int colIndex );

		/// return a Blob containing the specified data
		IBPP::Blob WriteBlob( BlobData bd, IBPP::Transaction m_tr );

		/// Execute a SQL string.
		/// The string is executed in a write transaction. If the database object has already
		/// a pending transaction then we will use that transaction. If there is no pending
		/// transaction we will use a temporary one for this command only.
		void ExecuteSQLString( const std::string &sql, const std::vector<BlobData> &bdv );

		/// load the vector with data from the blob fields in the row.
		void LoadBlobData( Variant *row, std::vector<BlobData> &bdv );

		std::string CreateSQLStringForInsert( Variant *row );
		std::string CreateSQLStringForUpdate( Variant *row );

	};

}

#endif

