/*
This file is part of FSDB: a higher level library to access a Firebird database using IBPP.

Copyright Fulvio Senore.

License:

Permission is hereby granted, free of charge, to any person or organization ("You") obtaining a copy of this software and 
associated documentation files covered by this license (the "Software") to use the Software as part of another work; 
to modify it for that purpose; to publish or distribute it, modified or not, for that same purpose; 
to permit persons to whom the other work using the Software is furnished to do so; subject to the following conditions: 
the above copyright notice and this complete and unmodified permission notice shall be included in all copies or substantial 
portions of the Software; You will not misrepresent modified versions of the Software as being the original.

The Software is provided "as is", without warranty of any kind, express or implied, including but not limited 
to the warranties of merchantability, fitness for a particular purpose and noninfringement. 
In no event shall the authors or copyright holders be liable for any claim, damages or other liability, 
whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software 
or the use of other dealings in the Software.
*/


#ifndef __FSDB_RECORDSET_BASE_H__
#define __FSDB_RECORDSET_BASE_H__

#include "variant.h"
#include "parameters.h"
#include "db_interface_base.h"

#include <string>

namespace FSDB {

	/// Abstract base class for recordset objects
	class RecordsetBase {

	public:

		RecordsetBase() : sql(""), DbInterface(NULL), isOpened(false) {}
		virtual ~RecordsetBase() {}
		
		/// open a recordset using a query without parameters
		virtual void Open( const std::string &sql, const std::string &tableName = "", const std::string &primaryKeyName = "" ) = 0;
		/// open a recordset using a query with parameters
		virtual void Open( const std::string &sql, Parameters params, const std::string &tableName = "", const std::string &primaryKeyName = "" ) = 0;

		virtual void Close() = 0;

		// return true if the recordset is opened
		bool IsOpened() { return isOpened; }

		/// return true if we moved past the last row (EOF)
		virtual bool IsEOF() = 0;
		/// return true if we moved before the first row (BOF)
		virtual bool IsBOF() = 0;
		/// return true if the recordset is empty
		virtual bool IsEmpty() = 0;

		virtual void MoveNext() = 0;
		virtual void MovePrevious() = 0;
		virtual void MoveFirst() = 0;
		virtual void MoveLast() = 0;

        // interface for random access to cached data
        // this will be implemented only in Recordset_FB_RW
        
        /// move the current position to the specified position
        /// valid newPosition values are those already read from the database
        /// a value not yet read from the database causes an exception
        virtual void MoveAbsolute( long newPosition ) = 0;
        /// return the current position in the recordset
        virtual long GetAbsolutePosition() = 0;
        /// save the corrent position for later use
        /// can be nested
        virtual void SaveAbsolutePosition() = 0;
        /// restore the previously saved position
        virtual void RestoreAbsolutePosition() = 0;
        /// return the number of already cached rows
        virtual long GetNumberOfCachedRows() = 0;
        /// return true if we have read all the rows from the database
        virtual bool HasReadAllRows() = 0;

        /// start editing mode for the current row
		virtual void Edit() = 0;
		/// start appending a new row
		virtual void AddNew() = 0;
		/// ends an Edit() or an AddNew() writing data to the database
		virtual void Update() = 0;
        /// cancels a pending Edit() or Update()
		virtual void CancelUpdate() = 0;


        /// delete the current row.
        /// the current record does not change, but it will not be accessible any more
        virtual void Delete() = 0;

		/// return a variant containing the value of a column in the current row
		virtual Variant& operator[]( const std::string& column ) = 0;
		virtual Variant& operator[]( const char* column ) { std::string s(column); return operator[](s); }

	protected:

		/// sql command used to create the recordset
		std::string sql;

		/// pointer to the database interface object
		DbInterfaceBase *DbInterface;

		/// type of the recordset
		RecordsetType rt;

		/// true if the recordset is opened
		bool isOpened;

		virtual void ReleaseMemory() = 0;

	};

	/// the type name used by programs for a pointer to a recordset object
	typedef boost::shared_ptr<RecordsetBase> RecordsetPt;
	/// the type name used by programs for a reference to a recordset object (used for a simpler syntax of the [] operator)
	typedef RecordsetBase& Recordset;

}

#endif

