/*
This file is part of FSDB: a higher level library to access a Firebird database using IBPP.

Copyright Fulvio Senore.

License:

Permission is hereby granted, free of charge, to any person or organization ("You") obtaining a copy of this software and 
associated documentation files covered by this license (the "Software") to use the Software as part of another work; 
to modify it for that purpose; to publish or distribute it, modified or not, for that same purpose; 
to permit persons to whom the other work using the Software is furnished to do so; subject to the following conditions: 
the above copyright notice and this complete and unmodified permission notice shall be included in all copies or substantial 
portions of the Software; You will not misrepresent modified versions of the Software as being the original.

The Software is provided "as is", without warranty of any kind, express or implied, including but not limited 
to the warranties of merchantability, fitness for a particular purpose and noninfringement. 
In no event shall the authors or copyright holders be liable for any claim, damages or other liability, 
whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software 
or the use of other dealings in the Software.
*/

#include "variant.h"

#include <assert.h>
#include <memory.h>
#include <stdexcept>


std::string FSDB::Variant::strBoolTrueValue = "T";
std::string FSDB::Variant::strBoolFalseValue = "F";


FSDB::Variant::VariantData::VariantData()
{
    m_CurrentType = Variant::ctNull;
}    

FSDB::Variant::VariantData::~VariantData() {
    ReleaseMemory();
}

void FSDB::Variant::VariantData::operator=(const FSDB::Variant::VariantData& v)
{
    Copy( v );
}


FSDB::Variant::VariantData::VariantData( const FSDB::Variant::VariantData& v )
{
    Copy( v );
}

void FSDB::Variant::VariantData::ReleaseMemory()
{
    if( m_CurrentType == FSDB::Variant::ctBlob ) {
		assert( d.blobData.pointer != NULL );
		delete static_cast<char *>( d.blobData.pointer );
	}
}

void FSDB::Variant::VariantData::Copy( const FSDB::Variant::VariantData& v )
{
    ReleaseMemory();

    m_CurrentType = v.m_CurrentType;
	switch( m_CurrentType ) {
		case FSDB::Variant::ctString:
			stringValue = v.stringValue;
			break;
		case FSDB::Variant::ctFloat:
			d.floatValue = v.d.floatValue;
			break;
		case FSDB::Variant::ctDouble:
			d.doubleValue = v.d.doubleValue;
			break;
		case FSDB::Variant::ctSmallInt:
			d.smallIntValue = v.d.smallIntValue;
			break;
		case FSDB::Variant::ctInteger:
			d.integerValue = v.d.integerValue;
			break;
		case FSDB::Variant::ctLargeInt:
			d.largeIntValue = v.d.largeIntValue;
			break;
		case FSDB::Variant::ctTimestamp:
			timestampValue = v.timestampValue;
			break;
		case FSDB::Variant::ctBlob:
			d.blobData.size = v.d.blobData.size;
			d.blobData.pointer = new char[d.blobData.size];
			memcpy( d.blobData.pointer, v.d.blobData.pointer, d.blobData.size );
			break;
	}
}

FSDB::Variant::Variant() {
    Init();
}

void FSDB::Variant::Init()
{
	changed = false;
    handleOrgValue = true;
}

void FSDB::Variant::SetAsFloat( float value ) {
    if( !(*this == value) )
        StoreOriginalValue();
    TestFreeMemory( ctFloat );
	curData.m_CurrentType = ctFloat;
	curData.d.floatValue = value;
}

void FSDB::Variant::SetAsDouble( double value ) {
    if( !(*this == value) )
        StoreOriginalValue();
	TestFreeMemory( ctDouble );
	curData.m_CurrentType = ctDouble;
	curData.d.doubleValue = value;
}

void FSDB::Variant::SetAsString( const std::string &value, bool isBoolean ) {
    bool newValueDifferent;
    if( isBoolean ) {
        std::string s = GetAsString();
        // remove trailing spaces
        while( s.size() > 0 && s[s.size() - 1] == ' ' )
            s = s.substr( 0, s.size() - 1 );
        newValueDifferent = (s != value);
    }
    else {
        newValueDifferent = !(*this == value);
    }
    if( newValueDifferent )
        StoreOriginalValue();
	TestFreeMemory( ctString );
	curData.m_CurrentType = ctString;
	curData.stringValue = value;
}

void FSDB::Variant::SetAsSmallInt( int16_t value ) {
    if( !(*this == value) )
        StoreOriginalValue();
	TestFreeMemory( ctSmallInt );
	curData.m_CurrentType = ctSmallInt;
	curData.d.smallIntValue = value;
}

void FSDB::Variant::SetAsInteger( int32_t value ) {
    if( !(*this == value) )
        StoreOriginalValue();
	TestFreeMemory( ctInteger );
	curData.m_CurrentType = ctInteger;
	curData.d.integerValue = value;
}

void FSDB::Variant::SetAsLargeInt( int64_t value ) {
    if( !(*this == value) )
        StoreOriginalValue();
	TestFreeMemory( ctLargeInt );
	curData.m_CurrentType = ctLargeInt;
	curData.d.largeIntValue = value;
}

void FSDB::Variant::SetAsTimestamp( const Timestamp &value ) {
    if( !(*this == value) )
        StoreOriginalValue();
	TestFreeMemory( ctTimestamp );
	curData.m_CurrentType = ctTimestamp;
	curData.timestampValue = value;
}

void FSDB::Variant::SetAsBool( bool value ) {
    if( value )
        SetAsString( strBoolTrueValue, true );
    else
        SetAsString( strBoolFalseValue, true );
}

void FSDB::Variant::SetAsBlobNoMemoryAllocation( void *pt, int size ) {
    bool different = (this->curData.m_CurrentType != ctBlob);
    if( !different ) different = (size != curData.d.blobData.size);
    if( !different ) different = (memcmp(pt, curData.d.blobData.pointer, size) != 0);
    if( different )
        StoreOriginalValue();
	TestFreeMemory( ctBlob );
	curData.m_CurrentType = ctBlob;
	curData.d.blobData.pointer = pt;
	curData.d.blobData.size = size;
}

void FSDB::Variant::SetAsBlob( void *pt, int size ) {
    bool different = (this->curData.m_CurrentType != ctBlob);
    if( !different ) different = (size != curData.d.blobData.size);
    if( !different ) different = (memcmp(pt, curData.d.blobData.pointer, size) != 0);
    if( different )
        StoreOriginalValue();
	TestFreeMemory( ctBlob );
	curData.m_CurrentType = ctBlob;
	curData.d.blobData.pointer = new char[size];
	memcpy( curData.d.blobData.pointer, pt, size );
	curData.d.blobData.size = size;
}

void FSDB::Variant::TestFreeMemory( Variant::CurrentType newType ) {

	if( curData.m_CurrentType == ctString && newType != ctString )
		curData.stringValue.erase();

	if( curData.m_CurrentType == ctBlob ) {
		assert( curData.d.blobData.pointer != NULL );
		delete static_cast<char *>( curData.d.blobData.pointer );
	}
}

void FSDB::Variant::SetNull() {
    StoreOriginalValue();
	TestFreeMemory( ctNull );
	curData.m_CurrentType = ctNull;
}


FSDB::Variant& FSDB::Variant::operator =( const std::string& value ) {
	SetAsString( value );
	return *this;
}

FSDB::Variant& FSDB::Variant::operator =( const float value ) {
	SetAsFloat( value );
	return *this;
}

FSDB::Variant& FSDB::Variant::operator =( const double value ) {
	SetAsDouble( value );
	return *this;
}

FSDB::Variant& FSDB::Variant::operator =( const int16_t value ) {
	SetAsSmallInt( value );
	return *this;
}

FSDB::Variant& FSDB::Variant::operator =( const int32_t value ) {
	SetAsInteger( value );
	return *this;
}

FSDB::Variant& FSDB::Variant::operator =( const int64_t value ) {
	SetAsLargeInt( value );
	return *this;
}

FSDB::Variant& FSDB::Variant::operator =( const Timestamp &value ) {
	SetAsTimestamp( value );
	return *this;
}

FSDB::Variant& FSDB::Variant::operator =( const bool value ) {
	SetAsBool( value );
	return *this;
}

std::string FSDB::Variant::GetAsString() const {
	TestCurrentType( ctString );
	return curData.stringValue;
}


float FSDB::Variant::GetAsFloat() const {
	float retVal = 0;

	TestCurrentType( ctFloat );
	switch( curData.m_CurrentType ) {
		case ctFloat:
			retVal = curData.d.floatValue;
			break;
		case ctDouble:
			retVal = (float) curData.d.doubleValue;
			break;
	}
	return retVal;
}

double FSDB::Variant::GetAsDouble() const {
	double retVal = 0;

	TestCurrentType( ctDouble );
	switch( curData.m_CurrentType ) {
		case ctFloat:
			retVal = curData.d.floatValue;
			break;
		case ctDouble:
			retVal = curData.d.doubleValue;
			break;
	}
	return retVal;
}

int16_t FSDB::Variant::GetAsSmallInt() const {
	int16_t retVal = 0;

	TestCurrentType( ctSmallInt );
	switch( curData.m_CurrentType ) {
		case ctSmallInt:
			retVal = curData.d.smallIntValue;
			break;
		case ctInteger:
			retVal = (int16_t) curData.d.integerValue;
			break;
		case ctLargeInt:
			retVal = (int16_t) curData.d.largeIntValue;
			break;
	}
	return retVal;
}


int32_t FSDB::Variant::GetAsInteger() const {
	int32_t retVal = 0;

	TestCurrentType( ctInteger );
	switch( curData.m_CurrentType ) {
		case ctSmallInt:
			retVal = curData.d.smallIntValue;
			break;
		case ctInteger:
			retVal = curData.d.integerValue;
			break;
		case ctLargeInt:
			retVal = (int32_t) curData.d.largeIntValue;
			break;
	}
	return retVal;
}


int64_t FSDB::Variant::GetAsLargeInt() const {
	int64_t retVal = 0;

	TestCurrentType( ctLargeInt );
	switch( curData.m_CurrentType ) {
		case ctSmallInt:
			retVal = curData.d.smallIntValue;
			break;
		case ctInteger:
			retVal = curData.d.integerValue;
			break;
		case ctLargeInt:
			retVal = curData.d.largeIntValue;
			break;
	}
	return retVal;
}

long FSDB::Variant::GetAsLong() const {
	long retVal = 0;

	TestCurrentType( ctInteger );
	switch( curData.m_CurrentType ) {
		case ctSmallInt:
			retVal = (long) curData.d.smallIntValue;
			break;
		case ctInteger:
			retVal = (long) curData.d.integerValue;
			break;
		case ctLargeInt:
			retVal = (long) curData.d.largeIntValue;
			break;
	}
	return retVal;
}


FSDB::Timestamp FSDB::Variant::GetAsTimestamp() const {
	TestCurrentType( ctTimestamp );
	return curData.timestampValue;
}

bool FSDB::Variant::GetAsBool() const {
    bool retVal;
    std::string s = GetAsString();
    
    // remove trailing spaces
    while( s.size() > 0 && s[s.size() - 1] == ' ' )
        s = s.substr( 0, s.size() - 1 );

    if( s == strBoolTrueValue )
        retVal = true;
    else {
        if( s == strBoolFalseValue )
            retVal = false;
        else
            throw std::runtime_error( "Variant::GetAsBool() called with a wrong internal string value" );
    }
    return retVal;
}

int FSDB::Variant::GetBlobSize() const {
	TestCurrentType( ctBlob );
	return curData.d.blobData.size;
}

void FSDB::Variant::GetAsBlob( void* &pt, int size ) const {
	TestCurrentType( ctBlob );
	if( size < curData.d.blobData.size )
		throw std::runtime_error( "The size parameter is too small for this blob data" );

	memcpy( pt, curData.d.blobData.pointer, curData.d.blobData.size );
}

void FSDB::Variant::GetAsBlobConstPointer( const void* &pt, int &size ) const {
	TestCurrentType( ctBlob );
	size = curData.d.blobData.size;
	pt = curData.d.blobData.pointer;
}

void FSDB::Variant::TestCurrentType( Variant::CurrentType testType ) const {
	bool failed = false;
	
	switch( curData.m_CurrentType ) {
		case ctNull:
			if( testType != ctString )
				throw std::runtime_error( "Attempt to convert a null variant to another type" );
			break;
		case ctString:
			if( testType != curData.m_CurrentType )
				failed = true;
			break;
		case ctFloat:
		case ctDouble:
			if( testType != ctFloat && testType != ctDouble )
				failed = true;
			break;
		case ctSmallInt:
		case ctInteger:
		case ctLargeInt:
			if( testType != ctSmallInt && testType != ctInteger && testType != ctLargeInt )
				failed = true;
			break;
		case ctTimestamp:
		case ctBlob:
			if( testType != curData.m_CurrentType )
				failed = true;
			break;
	}
		
	if( failed )
		throw std::runtime_error( "Attempt to read an incompatible data type from a Variant variable" );
}

void FSDB::Variant::operator=(const FSDB::Variant& v)
{
    Copy( v );
}


FSDB::Variant::Variant( const FSDB::Variant& v )
{
    Init();
    Copy( v );
}

void FSDB::Variant::Copy( const FSDB::Variant& v )
{
    ClearChanged();
    changed = v.changed;
    handleOrgValue = v.handleOrgValue;

    curData = v.curData;
    if( changed ) {
        orgData = v.orgData;
    }
}

void FSDB::Variant::StoreOriginalValue()
{
    if( handleOrgValue && !changed ) {
        // keep a copy of the original value
        orgData = curData;
    }
   	changed = true;
}

void FSDB::Variant::RestoreOriginalValue()
{
    if( !handleOrgValue ) return;
    if( !changed ) return;

    curData = orgData;
    ClearChanged();
}

void FSDB::Variant::ClearChanged()
{
    changed = false;
}

int FSDB::Variant::operator==( const int16_t right ) 
{ 
    if( this->curData.m_CurrentType != ctSmallInt && this->curData.m_CurrentType != ctInteger && this->curData.m_CurrentType != ctInteger ) return false;
    return (this->GetAsSmallInt() == right); 
}

int FSDB::Variant::operator==( const int32_t right )
{
    if( this->curData.m_CurrentType != ctInteger && this->curData.m_CurrentType != ctLargeInt && this->curData.m_CurrentType != ctSmallInt )  return false;
    return (this->GetAsInteger() == right);
}

int FSDB::Variant::operator==( const int64_t right )
{
    if( this->curData.m_CurrentType != ctLargeInt && this->curData.m_CurrentType != ctInteger && this->curData.m_CurrentType != ctSmallInt ) return false;
    return (this->GetAsLargeInt() == right);
}

int FSDB::Variant::operator==( const std::string& right )
{
    if( this->curData.m_CurrentType != ctString ) return false;
    return (this->GetAsString() == right);
}

int FSDB::Variant::operator==( const float right )
{
    if( this->curData.m_CurrentType != ctFloat && this->curData.m_CurrentType != ctDouble ) return false;
    return (this->GetAsFloat() == right);
}

int FSDB::Variant::operator==( const double right )
{
    if( this->curData.m_CurrentType != ctDouble && this->curData.m_CurrentType != ctFloat ) return false;
    return (this->GetAsDouble() == right);
}

int FSDB::Variant::operator==( const Timestamp &right )
{
    if( this->curData.m_CurrentType != ctTimestamp ) return false;
    return (this->GetAsTimestamp() == right);
}

int FSDB::Variant::operator==( const bool right )
{
    if( this->curData.m_CurrentType != ctString ) return false;
    return (this->GetAsBool() == right);
}

int FSDB::Variant::operator==( const Variant &right )
{
    bool r = false;

    if( this->GetCurrentType() != right.GetCurrentType() ) return false;

	// here if both variants have the same type
    switch( curData.m_CurrentType ) {
		case ctNull:
            r = true;
            break;
        case ctString:
            r = (this->GetAsString() == right.GetAsString());
            break;
        case ctFloat:
            r = (this->GetAsFloat() == right.GetAsFloat());
            break;
        case ctDouble:
            r = (this->GetAsDouble() == right.GetAsDouble());
            break;
		case ctSmallInt:
		case ctInteger:
		case ctLargeInt:
            r = (this->GetAsLargeInt() == right.GetAsLargeInt());
            break;
		case ctTimestamp:
            r = (this->GetAsTimestamp() == right.GetAsTimestamp());
            break;
        default:
            // not supported
            assert( false );
    }
    return r;
}

int FSDB::Variant::operator!=( const Variant &right )
{
    return !operator==( right );
}


#ifdef _WX_FSDB_

// first some conversion functions

//-----------------------------------------------------------------------------
//! converts wxString to std::string
std::string wx2std(const wxString& input, wxMBConv* conv)
{
    if (input.empty())
        return "";
    if (!conv)
        conv = wxConvCurrent;
    const wxWX2MBbuf buf(input.mb_str(*conv));
    // conversion may fail and return 0, which isn't a safe value to pass 
    // to std:string constructor
    if (!buf)
        return "";
    return std::string(buf);
}
//-----------------------------------------------------------------------------
//! converts std:string to wxString
wxString std2wx(const std::string& input, wxMBConv* conv)
{
    if (input.empty())
        return wxEmptyString;
    if (!conv)
        conv = wxConvCurrent;
    return wxString(input.c_str(), *conv);
}

// string conversion for database access
// the database uses UTF8 encoding
std::string DBwx2std( const wxString& input ) {
	std::string s = wx2std( input, &wxConvUTF8 );
	return wx2std( input, &wxConvUTF8 );
}
wxString DBstd2wx( const std::string& input ) {
	return std2wx( input, &wxConvUTF8 );
}

FSDB::Timestamp dt_wx2fsdb( const wxDateTime& dt ) {
	return FSDB::Timestamp( dt.GetYear(), ((int) dt.GetMonth()) + 1, dt.GetDay(), dt.GetHour(), dt.GetMinute(), dt.GetSecond() );
}

wxDateTime dt_fsdb2wx( const FSDB::Timestamp& ts ) {
	return wxDateTime( ts.Day(), (wxDateTime::Month) (ts.Month() - 1), ts.Year(), ts.Hours(), ts.Minutes(), ts.Seconds() );
}

// now the wxWidgets related methods

void FSDB::Variant::SetAswxString( const wxString &value )
{
    std::string stmp = DBwx2std( value );
    SetAsString( stmp );
}

void FSDB::Variant::SetAswxDateTime( const wxDateTime &value )
{
    if( value.IsValid() ) {
        FSDB::Timestamp ts = dt_wx2fsdb( value );
        SetAsTimestamp( ts );
    }
    else
        SetNull();
}

wxString FSDB::Variant::GetAswxString() const
{
    std::string stmp = GetAsString();
    return DBstd2wx( stmp );
}

wxDateTime FSDB::Variant::GetAswxDateTime() const
{
    wxDateTime dt;
    if( IsNull() )
        dt = wxInvalidDateTime;
    else {
        FSDB::Timestamp ts = GetAsTimestamp();
        dt = dt_fsdb2wx( ts );
    }
    
    return dt;
}

FSDB::Variant& FSDB::Variant::operator=( const wxString& value )
{
	SetAswxString( value );
	return *this;
}

FSDB::Variant& FSDB::Variant::operator=( const wxDateTime &value )
{
    SetAswxDateTime( value );
	return *this;
}

int FSDB::Variant::operator==( const wxString& right )
{
    if( this->curData.m_CurrentType != ctString ) return false;
    return (this->GetAswxString() == right);
}

int FSDB::Variant::operator==( const wxDateTime &right )
{
    if( this->curData.m_CurrentType != ctTimestamp ) return false;
    return (this->GetAswxDateTime() == right);
}




#endif  // _WX_FSDB_


