/*
This file is part of FSDB: a higher level library to access a Firebird database using IBPP.

Copyright Fulvio Senore.

License:

Permission is hereby granted, free of charge, to any person or organization ("You") obtaining a copy of this software and 
associated documentation files covered by this license (the "Software") to use the Software as part of another work; 
to modify it for that purpose; to publish or distribute it, modified or not, for that same purpose; 
to permit persons to whom the other work using the Software is furnished to do so; subject to the following conditions: 
the above copyright notice and this complete and unmodified permission notice shall be included in all copies or substantial 
portions of the Software; You will not misrepresent modified versions of the Software as being the original.

The Software is provided "as is", without warranty of any kind, express or implied, including but not limited 
to the warranties of merchantability, fitness for a particular purpose and noninfringement. 
In no event shall the authors or copyright holders be liable for any claim, damages or other liability, 
whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software 
or the use of other dealings in the Software.
*/

/*
Handling of boolean values:

many databases do not have a native boolean type, so the Variant class represent those values as predefined strings,
one for the true value and one for the false value. The database sees them as strings, and the Variant class has methods to
map those strings to the boolean values.

*/

#ifndef __FSDB_VARIANT_H__
#define __FSDB_VARIANT_H__

#include <string>

#ifdef _WX_FSDB_
  #include <wx/string.h>
  #include <wx/datetime.h>
  #include <wx/strconv.h>
#endif  // _WX_FSDB_

#include "date_time.h"
#include "_fsdb.h"


namespace FSDB {



	/// a class that can hold data of different types
	class Variant {

	public:

		/// the data type that this object is currently holding
		enum CurrentType { 
			ctNull, /*!< null */
			ctString, /*!< string */
			ctFloat,  /*!< float */
			ctDouble,  /*!< double */
			ctSmallInt,  /*!< small integer */
			ctInteger,  /*!< integer */
			ctLargeInt,  /*!< large integer */
			ctTimestamp,  /*!< timestamp */
			ctBlob   /*!< blob */
		};

		Variant();
		Variant( float value ) { SetAsFloat(value); }
		Variant( double value ) { SetAsDouble(value); }
		Variant( const std::string &value ) { SetAsString(value); }
		Variant( int16_t value ) { SetAsSmallInt(value); }
		Variant( int32_t value ) { SetAsInteger(value); }
		Variant( int64_t value ) { SetAsLargeInt(value); }
		Variant( const Timestamp &value ) { SetAsTimestamp(value); }
		Variant( bool value ) { SetAsBool(value); }

		/// copy constructor
		Variant( const Variant& v );
        /// assignment operator
    	void operator=(const Variant& v);
        /// copy data from another instance
        void Copy(const Variant& v);

		void SetAsFloat( float value );
		void SetAsDouble( double value );
		// if isBoolean == true the input string represents a boolean value 
        // so when deciding if the current value has changed trailing spaces will be ignored
        void SetAsString( const std::string &value, bool isBoolean = false );
		void SetAsSmallInt( int16_t value );
		void SetAsInteger( int32_t value );
		void SetAsLargeInt( int64_t value );
		void SetAsTimestamp( const Timestamp &value );
		void SetAsBool( bool value );
		/// set data as a blob.
		/// pt is the pointer to the data, size is the data size
		/// the Variant will make a private copy of the data
		void SetAsBlob( void *pt, int size );
		/// set data as a blob.
		/// pt is the pointer to the data, size is the data size
		/// the Variant will become the owner of the memory block and it will delete it when needed
		/// the caller MUST NOT delete the memory block
		void SetAsBlobNoMemoryAllocation( void *pt, int size );

		std::string GetAsString() const;
		float GetAsFloat() const;
		double GetAsDouble() const;
		int16_t GetAsSmallInt() const;
		int32_t GetAsInteger() const;
		int64_t GetAsLargeInt() const;
		long GetAsLong() const;
		Timestamp GetAsTimestamp() const;
		bool GetAsBool() const;
		/// return the size of the blob data
		int GetBlobSize() const;
		/// return a copy of the blob data.
		/// copies data to the pointer, size is the size of the memory allocated to the pointer
        void GetAsBlob( void* &pt, int size ) const;
		/// return a const pointer the the blob data.
		/// size is the size of the memory allocated to the pointer
		/// this function is faster than GetAsBlob() because it does not copy the data
		/// but the user cannot modify the data because they see he internal Variant's data
        void GetAsBlobConstPointer( const void* &pt, int &size ) const;

		Variant& operator=( const std::string& value );
		Variant& operator=( const float value );
		Variant& operator=( const double value );
		Variant& operator=( const int16_t value );
		Variant& operator=( const int32_t value );
		Variant& operator=( const int64_t value );
		Variant& operator=( const Timestamp &value );
		Variant& operator=( const bool value );

		int operator==( const int16_t right );
		int operator==( const int32_t right );
		int operator==( const int64_t right );
		int operator==( const std::string& right );
		int operator==( const float right );
		int operator==( const double right );
		int operator==( const bool right );
		int operator==( const Timestamp &right );

        int operator==( const Variant &right );
        int operator!=( const Variant &right );

		operator std::string() { return GetAsString(); }
		operator float() { return GetAsFloat(); }
		operator double() { return GetAsDouble(); }
		operator int16_t() { return GetAsSmallInt(); }
		operator int32_t() { return GetAsInteger(); }
		operator int64_t() { return GetAsLargeInt(); }
		operator long() { return GetAsLong(); }
		operator Timestamp() { return GetAsTimestamp(); }
		operator bool() { return GetAsBool(); }

		bool IsNull() const { return (curData.m_CurrentType == ctNull); }
		
		/// set this instance to NULL
		void SetNull();

		/// set the changed flag to false
		void ClearChanged();

		/// true if the value has been changed
		bool IsChanged() const { return changed; }

		/// return the type contained in the variant
		CurrentType GetCurrentType() const { return curData.m_CurrentType; }

		/// if true this instance will store its original value if changed
        void SetHandleOriginalValue( bool doHandleOriginalValue ) { handleOrgValue = doHandleOriginalValue; }

        bool GetHandleOriginalValue() const { return handleOrgValue; }

        /// restore the original value of this instance.
        /// nothing happens if this instance has not changed
        void RestoreOriginalValue();

#ifdef _WX_FSDB_
    // interface with wxWidgets

		Variant( const wxString &value ) { SetAswxString(value); }
		Variant( const wxDateTime &value ) { SetAswxDateTime(value); }

        void SetAswxString( const wxString &value );
		void SetAswxDateTime( const wxDateTime &value );    // wxInvalidDateTime is converted to NULL

		wxString GetAswxString() const;
		wxDateTime GetAswxDateTime() const;         // NULL is converted to wxInvalidDateTime

		Variant& operator=( const wxString& value );
		Variant& operator=( const wxDateTime &value );

		int operator==( const wxString& right );
		int operator==( const wxDateTime &right );

		operator wxString() const { return GetAswxString(); }
		operator wxDateTime() const { return GetAswxDateTime(); }
#endif  // _WX_FSDB_

    protected:

        /// used by the Variant class, holds actual data
        class VariantData
        {
        public:

            VariantData();
		    ~VariantData();

		    /// copy constructor
		    VariantData( const VariantData& v );
            /// assignment operator
    	    void operator=(const VariantData& v);
            /// copy data from another instance
            void Copy(const VariantData& v);

            /// release the memory used by this instance (if any)
            void ReleaseMemory();

		    /// the type currently contained in this instance
            Variant::CurrentType m_CurrentType;

            /// a union that shares data for some of the possible types
		    union data {
			    float floatValue; /*!< the float value */
			    double doubleValue; /*!< the double value */
			    int16_t smallIntValue; /*!< the small integer value */
			    int32_t integerValue; /*!< the integer value */
			    int64_t largeIntValue; /*!< the large integer value */
			    struct {
				    void *pointer;	/*!< address of the blob */
				    int size;		/*!< size of the blob */
			    } blobData;	/*!< data used for blob values, handled as a void pointer */
		    } d;
		    /// holds the string value
		    std::string stringValue;
		    /// holds the timestamp value
		    Timestamp timestampValue;
        };


        /// the actual current data
        VariantData curData;

        /// the original data, if any.
        /// these data are only meningful if changed == true
        VariantData orgData;

		/// flag, true if the value has been changed
		bool changed;

        /// if true this instance will store its original value if changed.
        /// set to false in the instance that keeps the original value to avoid recursion
        bool handleOrgValue;

        // static strings that contain the string representation for boolean values
        static std::string strBoolTrueValue;
        static std::string strBoolFalseValue;

        /// store the original value if this is the first change to this instance
        void StoreOriginalValue();

    		/// check the type of the current value.
		/// checks that the current type stored by this instance
		/// is the same as the parameter: if it is not the same (or compatible) it throws an exception
		void TestCurrentType( CurrentType testType ) const;

		/// check if we need to free some memory.
		/// called when assigning a new value, newType is the type of the new value
		void TestFreeMemory( CurrentType newType );

        /// initialization of a new object
        void Init();
    };

}

#endif
