/*
	This file is part of vvvP (Virtual Volumes View Photo edition)

	Copyright (C) 2008, the vvvP Development team

	Author: Fulvio Senore

    vvvP is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    vvvP is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with vvvP; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

// ----------------------------------------------------------------------------
// headers
// ----------------------------------------------------------------------------

#include "wx/wxprec.h"

#ifdef __BORLANDC__
    #pragma hdrstop
#endif

#ifndef WX_PRECOMP
    #include "wx/textctrl.h"
    #include "wx/dcclient.h"
#endif

#include "date_picker.h"

#include "wx/settings.h"

// ----------------------------------------------------------------------------
// utility functions
// ----------------------------------------------------------------------------

// return a date one day later
static wxDateTime DateAddDay( wxDateTime &dt )
{
    wxDateSpan dsp(0, 0, 0, 1);
    wxDateTime retVal = dt + dsp;
    return retVal;
}

// return a date one day earlier
static wxDateTime DateSubtractDay( wxDateTime &dt )
{
    wxDateSpan dsp(0, 0, 0, 1);
    wxDateTime retVal = dt - dsp;
    return retVal;
}

// return a date one datespan earlier of later
// if forward == true go forward in time, else go backwards
// if the input date is end of month also the resulting date will be end of month
// this function is designed to add/subtract one month or one year
static wxDateTime DateAddSpan( wxDateTime &dt, wxDateSpan dsp, bool forward )
{
    wxDateSpan dspMonth(0, 1, 0, 0);
    wxDateSpan dspDay(0, 0, 0, 1);
    wxDateTime retVal = dt;

    // see if input is at the end of a month
    bool eom = (dt == retVal.SetToLastMonthDay());

    // move by a month
    if( forward )
        retVal = dt + dsp;
    else
        retVal = dt - dsp;

    if( eom ) {
        // make sure that the result is at the end of the new month
        retVal = retVal + dspMonth;
        retVal.SetDay( 1 );
        retVal = retVal - dspDay;
    }

    return retVal;
}



// return a date one month earlier of later
// if forward == true go forward in time, else go backwards
// if the input date is end of month also the resulting date will be end of month
static wxDateTime DateAddMonth( wxDateTime &dt, bool forward )
{
    wxDateSpan dspMonth(0, 1, 0, 0);

    return DateAddSpan( dt, dspMonth, forward );
}

// return a date one year earlier of later
// if forward == true go forward in time, else go backwards
// if the input date is end of month also the resulting date will be end of month
static wxDateTime DateAddYear( wxDateTime &dt, bool forward )
{
    wxDateSpan dspYear(1, 0, 0, 0);

    return DateAddSpan( dt, dspYear, forward );
}

// takes as input a string that should contain a date
// returns a string reformatted adding some default behaviours:
// - add separators if missing
// - add century digits if missing
static wxString FormatDateText( const wxString &orgText, const wxChar &separator )
{
    size_t k;

    if( orgText.length() != 6 && orgText.length() != 8 ) return orgText;

    wxString retVal = orgText;

    // count separators
    int nSeparators = 0;
    for( k = 0; k < retVal.length(); k++ ) {
        if( retVal[k] == separator ) {
            nSeparators++;
            if( k != 2 && k != 5 ) return orgText;  // we want 2 digits for day and 2 for month
        }
    }

    if( nSeparators != 0 && nSeparators != 2 ) return orgText;

    // add separators if missing
    if( nSeparators == 0 )
        retVal = retVal.Left(2) + separator + retVal.Mid(2, 2) + separator + retVal.Mid(4);

    // check if the year has only two digits
    if( retVal.length() == 8 )
        retVal = retVal.Left(6) + wxT("20") + retVal.Mid(6);

    return retVal;

}

// curValue is the current value of the text control
static void HandleOnChar(wxKeyEvent &event, const wxChar &separator, const wxString &curValue )
{
    bool ok = false;

    int keyCode = event.GetKeyCode();

	if( keyCode < WXK_SPACE || keyCode == WXK_DELETE || keyCode > WXK_START ) {
		// we don't filter special keys and Delete
		event.Skip();
		return;
	}

    if( keyCode >= '0' && keyCode <= '9' ) ok = true;

    if( keyCode == separator ) {
        // do not allow more than two separators
        int nSeparators = 0;
        for( size_t k = 0; k < curValue.length(); k ++ )
            if( curValue[k] == separator )
                nSeparators++;
        ok = (nSeparators < 2);
    }

    if( ok )
        event.Skip();
}

// handles the key down event for the date controls
// dt is an I/O parameter
// in input it is the current date in the control
// if the function returns false the date has not changed
// if the function returns true the new date is contained in the dt parameter
static bool HandleOnKeyDown(wxKeyEvent &event, wxDateTime &dt)
{
    int keyCode = event.GetKeyCode();

    if( keyCode == '+' || keyCode == WXK_ADD || keyCode == WXK_UP || keyCode == WXK_NUMPAD_ADD ) {
        if( dt.IsValid() ) {
            if( event.GetModifiers() == wxMOD_NONE )
                dt = DateAddDay( dt );
            if( event.GetModifiers() == wxMOD_CMD )
                dt = DateAddMonth( dt, true );
            if( event.GetModifiers() == wxMOD_ALT )
                dt = DateAddYear( dt, true );
            return true;
        }
    }
    if( keyCode == '-' || keyCode == WXK_SUBTRACT || keyCode == WXK_DOWN || keyCode == WXK_NUMPAD_SUBTRACT ) {
        if( dt.IsValid() ) {
            if( event.GetModifiers() == wxMOD_NONE )
                dt = DateSubtractDay( dt );
            if( event.GetModifiers() == wxMOD_CMD )
                dt = DateAddMonth( dt, false );
            if( event.GetModifiers() == wxMOD_ALT )
                dt = DateAddYear( dt, false );
            return true;
        }
    }

    event.Skip();

    return false;
}



// ----------------------------------------------------------------------------
// local classes
// ----------------------------------------------------------------------------

class fsCalendarComboPopup : public wxCalendarCtrl,
                             public wxComboPopup
{
public:

    fsCalendarComboPopup() : wxCalendarCtrl(),
                             wxComboPopup()
    {
    }

    virtual void Init()
    {
    }

    // NB: Don't create lazily since it didn't work that way before
    //     wxComboCtrl was used, and changing behaviour would almost
    //     certainly introduce new bugs.
    virtual bool Create(wxWindow* parent);
    virtual wxSize GetAdjustedSize(int minWidth, int prefHeight, int maxHeight);

    virtual wxWindow *GetControl() { return this; }

    void SetDateValue(const wxDateTime& date);

    bool IsTextEmpty() const;

    bool ParseDateTime(const wxString& s, wxDateTime* pDt);
    void SendDateEvent(const wxDateTime& dt);

    wxChar GetSeparator() { return m_separator; }
    bool HasValidSeparator() { return m_separator != m_invalidSeparator; }

private:

    // extract the date separator from a string containing a date
    void SetSeparator( wxString &sdate );

    void OnCalKey(wxKeyEvent & ev);
    void OnSelChange(wxCalendarEvent &ev);
    void OnKillTextFocus(wxFocusEvent &ev);

    bool HasDPFlag(int flag) const;

    // it expands "%x" format and changes %y to %Y if wxDP_SHOWCENTURY flag
    // is given. If the locale format can't be easily analyzed (e.g. when
    // the month is given as a name, not number), "%x" is returned
    wxString GetLocaleDateFormat();

    bool SetFormat(const wxString& fmt);

    virtual void SetStringValue(const wxString& s);
    virtual wxString GetStringValue() const;

    // returns either the given date representation using the current format or
    // an empty string if it's invalid
    wxString GetStringValueFor(const wxDateTime& dt) const;

    wxSize          m_useSize;
    wxString        m_format;
    wxChar          m_separator;  // the current date separator
    static const wxChar    m_invalidSeparator = '1';

    DECLARE_EVENT_TABLE()
};

bool fsCalendarComboPopup::Create(wxWindow* parent)
{
    if ( !wxCalendarCtrl::Create(parent, wxID_ANY, wxDefaultDateTime,
                            wxPoint(0, 0), wxDefaultSize,
                            wxCAL_SEQUENTIAL_MONTH_SELECTION
                            | wxCAL_SHOW_HOLIDAYS | wxBORDER_THEME) )
        return false;

    SetFormat(GetLocaleDateFormat());

    m_useSize  = wxCalendarCtrl::GetBestSize();

    wxWindow* tx = m_combo->GetTextCtrl();
    if ( !tx )
        tx = m_combo;

    tx->Connect(wxEVT_KILL_FOCUS,
                wxFocusEventHandler(fsCalendarComboPopup::OnKillTextFocus),
                NULL, this);

    return true;
}

void fsCalendarComboPopup::SetDateValue(const wxDateTime& date)
{
    if ( date.IsValid() )
    {
        m_combo->SetText(date.Format(m_format));
        SetDate(date);
    }
    else // invalid date
    {
        m_combo->SetText(wxEmptyString);
    }
}

bool fsCalendarComboPopup::ParseDateTime(const wxString& s, wxDateTime* pDt)
{
    wxASSERT(pDt);

    if ( !s.empty() )
    {
        pDt->ParseFormat(s, m_format);
        if ( !pDt->IsValid() )
            return false;
    }

    return true;
}

void fsCalendarComboPopup::SendDateEvent(const wxDateTime& dt)
{
    // Sends both wxCalendarEvent and wxDateEvent
    wxWindow* datePicker = m_combo->GetParent();

    wxCalendarEvent cev(datePicker, dt, wxEVT_CALENDAR_SEL_CHANGED);
    datePicker->GetEventHandler()->ProcessEvent(cev);

    wxDateEvent event(datePicker, dt, wxEVT_DATE_CHANGED);
    datePicker->GetEventHandler()->ProcessEvent(event);
}

void fsCalendarComboPopup::SetSeparator( wxString &sdate )
{
    m_separator = m_invalidSeparator;
    for( size_t k = 0; k < sdate.length(); k++ ) {
        if( !wxIsdigit(sdate[k]) ) {
            m_separator = sdate[k];
            break;
        }
    }
}


void fsCalendarComboPopup::OnCalKey(wxKeyEvent & ev)
{
    if (ev.GetKeyCode() == WXK_ESCAPE && !ev.HasModifiers())
        Dismiss();
    else
        ev.Skip();
}

void fsCalendarComboPopup::OnSelChange(wxCalendarEvent &ev)
{
    m_combo->SetText(GetDate().Format(m_format));

    if ( ev.GetEventType() == wxEVT_CALENDAR_DOUBLECLICKED )
    {
        Dismiss();
    }

    SendDateEvent(GetDate());
}

void fsCalendarComboPopup::OnKillTextFocus(wxFocusEvent &ev)
{
    ev.Skip();

    const wxDateTime& dtOld = GetDate();

    wxDateTime dt;
    wxString value = m_combo->GetValue();

    // add some default formatting
    value = FormatDateText( value, m_separator );

    if ( !ParseDateTime(value, &dt) )
    {
        dt = dtOld;
    }

    m_combo->SetText(GetStringValueFor(dt));

    if ( !dt.IsValid() )
        return;

    // notify that we had to change the date after validation
    if ( (dt.IsValid() && (!dtOld.IsValid() || dt != dtOld)) ||
            (!dt.IsValid() && dtOld.IsValid()) )
    {
        SetDate(dt);
        SendDateEvent(dt);
    }
}

wxString fsCalendarComboPopup::GetLocaleDateFormat()
{
    wxString fmt = wxLocale::GetInfo(wxLOCALE_SHORT_DATE_FMT);
    if ( HasDPFlag(wxDP_SHOWCENTURY) )
        fmt.Replace("%y", "%Y");

    // determine the date separator
    wxDateTime dt;
    dt.ParseFormat(wxT("2003-10-17"), wxT("%Y-%m-%d"));
    wxString str(dt.Format("%x"));

    SetSeparator( str );

    return fmt;
}

bool fsCalendarComboPopup::SetFormat(const wxString& fmt)
{
    m_format = fmt;

    if ( m_combo )
    {

        if ( GetDate().IsValid() )
            m_combo->SetText(GetDate().Format(m_format));
    }

    return true;
}

void fsCalendarComboPopup::SetStringValue(const wxString& s)
{
    wxDateTime dt;
    if ( !s.empty() && ParseDateTime(s, &dt) )
        SetDate(dt);
    //else: keep the old value
}

wxString fsCalendarComboPopup::GetStringValue() const
{
    return GetStringValueFor(GetDate());
}

wxString fsCalendarComboPopup::GetStringValueFor(const wxDateTime& dt) const
{
    wxString val;
    if ( dt.IsValid() )
        val = dt.Format(m_format);

    return val;
}

bool fsCalendarComboPopup::IsTextEmpty() const
{
    return m_combo->GetTextCtrl()->IsEmpty();
}

wxSize fsCalendarComboPopup::GetAdjustedSize(int WXUNUSED(minWidth),
                                int WXUNUSED(prefHeight),
                                int WXUNUSED(maxHeight))
{
    return m_useSize;
}

bool fsCalendarComboPopup::HasDPFlag(int flag) const
{
    return m_combo->GetParent()->HasFlag(flag);
}








BEGIN_EVENT_TABLE(fsCalendarComboPopup, wxCalendarCtrl)
    EVT_KEY_DOWN(fsCalendarComboPopup::OnCalKey)
    EVT_CALENDAR_SEL_CHANGED(wxID_ANY, fsCalendarComboPopup::OnSelChange)
    EVT_CALENDAR_PAGE_CHANGED(wxID_ANY, fsCalendarComboPopup::OnSelChange)
    EVT_CALENDAR(wxID_ANY, fsCalendarComboPopup::OnSelChange)
END_EVENT_TABLE()


// ============================================================================
// fsDatePickerCtrl implementation
// ============================================================================

BEGIN_EVENT_TABLE(fsDatePickerCtrl, wxComboCtrl)
    EVT_TEXT(wxID_ANY, fsDatePickerCtrl::OnText)
END_EVENT_TABLE()

IMPLEMENT_DYNAMIC_CLASS(fsDatePickerCtrl, wxComboCtrl)

// ----------------------------------------------------------------------------
// creation
// ----------------------------------------------------------------------------

bool fsDatePickerCtrl::Create(wxWindow *parent,
                                     wxWindowID id,
                                     const wxDateTime& date,
                                     const wxPoint& pos,
                                     const wxSize& size,
                                     long style,
                                     const wxValidator& validator)
{

    if ( !wxComboCtrl::Create(parent, id, wxEmptyString, pos, size,
                            style | wxCLIP_CHILDREN | wxWANTS_CHARS,
                            validator) )
    {
        return false;
    }

    m_popup = new fsCalendarComboPopup();

#if defined(__WXMSW__)
    // without this keyboard navigation in month control doesn't work
    UseAltPopupWindow();
#endif
    SetPopupControl(m_popup);

    m_popup->SetDateValue(date.IsValid() ? date : wxDateTime::Today());

    SetInitialSize(size);

    wxWindow* tx = GetTextCtrl();
    if ( !tx )
        tx = this;
    tx->Connect(wxEVT_CHAR,
                wxKeyEventHandler(fsDatePickerCtrl::OnChar),
                NULL, this);
    tx->Connect(wxEVT_KEY_DOWN,
                wxKeyEventHandler(fsDatePickerCtrl::OnKeyDown),
                NULL, this);

    return true;
}


void fsDatePickerCtrl::Init()
{
    m_popup = NULL;
}

fsDatePickerCtrl::~fsDatePickerCtrl()
{
}

bool fsDatePickerCtrl::Destroy()
{
    m_popup = NULL;

    return wxComboCtrl::Destroy();
}

// ----------------------------------------------------------------------------
// overridden base class methods
// ----------------------------------------------------------------------------

wxSize fsDatePickerCtrl::DoGetBestSize() const
{
    wxClientDC dc(const_cast<fsDatePickerCtrl *>(this));
    
    // sample string date: it contains more char to have more space
    wxString sep = "/";
    if( m_popup != NULL ) sep = m_popup->GetSeparator();
    wxString sdt = wxString(wxT("07")) + sep + wxT("05") + sep + wxT("201000");
    
    int x, y;
    dc.GetTextExtent(sdt, &x, &y);

    // account for the drop-down arrow or spin arrows
    x += wxSystemSettings::GetMetric( wxSYS_VSCROLL_X, const_cast<fsDatePickerCtrl *>(this) );
#ifndef __WXMSW__
	x += 6;		// some more room for GTK and Mac
#endif

    wxSize best = wxComboCtrl::DoGetBestSize();
    best.SetWidth( x );
    CacheBestSize(best);
    return best;
}

wxDateTime fsDatePickerCtrl::GetValueDT() const
{
    if ( m_popup->IsTextEmpty() )
        return wxInvalidDateTime;
    return m_popup->GetDate();
}

void fsDatePickerCtrl::SetValueDT(const wxDateTime& date)
{
    m_popup->SetDateValue(date);
}


// ----------------------------------------------------------------------------
// event handlers
// ----------------------------------------------------------------------------


void fsDatePickerCtrl::OnText(wxCommandEvent &ev)
{
    ev.SetEventObject(this);
    ev.SetId(GetId());
    GetParent()->GetEventHandler()->ProcessEvent(ev);

    // We'll create an additional event if the date is valid.
    // If the date isn't valid, the user's probably in the middle of typing
    wxDateTime dt;
    if ( !m_popup || !m_popup->ParseDateTime(GetValue(), &dt) )
        return;

    m_popup->SendDateEvent(dt);
}


void fsDatePickerCtrl::OnChar(wxKeyEvent &event)
{
    HandleOnChar( event, m_popup->GetSeparator(), GetValue() );
}

void fsDatePickerCtrl::OnKeyDown(wxKeyEvent &event)
{
    wxDateTime dt = GetValueDT();
    if( HandleOnKeyDown(event, dt) ) {
        SetValueDT( dt );
    }
}



// ============================================================================
// fsDatePickerTextCtrl implementation
// ============================================================================

BEGIN_EVENT_TABLE(fsDatePickerTextCtrl, wxTextCtrl)
    EVT_CHAR(fsDatePickerTextCtrl::OnChar)
    EVT_KEY_DOWN(fsDatePickerTextCtrl::OnKeyDown)
    EVT_KILL_FOCUS(fsDatePickerTextCtrl::OnKillFocus)
END_EVENT_TABLE()

IMPLEMENT_DYNAMIC_CLASS(fsDatePickerTextCtrl, wxTextCtrl)


fsDatePickerTextCtrl::~fsDatePickerTextCtrl()
{
}

bool fsDatePickerTextCtrl::Create(wxWindow *parent,
                                     wxWindowID id,
                                     const wxDateTime& date,
                                     const wxPoint& pos,
                                     const wxSize& size,
                                     long style,
                                     const wxValidator& validator)
{

    if ( !wxTextCtrl::Create(parent, id, wxEmptyString, pos, size,
                            style, validator) )
    {
        return false;
    }

    SetValueDT(date.IsValid() ? date : wxDateTime::Today());

    return true;
}

void fsDatePickerTextCtrl::Init()
{
    SetFormat(GetLocaleDateFormat());
}

wxString fsDatePickerTextCtrl::GetLocaleDateFormat()
{
    wxString fmt = wxLocale::GetInfo(wxLOCALE_SHORT_DATE_FMT);
    fmt.Replace("%y", "%Y");

    // determine the date separator
    wxDateTime dt;
    dt.ParseFormat(wxT("2003-10-17"), wxT("%Y-%m-%d"));
    wxString str(dt.Format("%x"));

    SetSeparator( str );

    return fmt;
}

void fsDatePickerTextCtrl::SetFormat(const wxString& fmt)
{
    m_format = fmt;
    SetValueDT( GetValueDT() );
}

void fsDatePickerTextCtrl::SetSeparator( wxString &sdate )
{
    m_separator = m_invalidSeparator;
    for( size_t k = 0; k < sdate.length(); k++ ) {
        if( !wxIsdigit(sdate[k]) ) {
            m_separator = sdate[k];
            break;
        }
    }
}

wxSize fsDatePickerTextCtrl::DoGetBestSize() const
{
    wxClientDC dc(const_cast<fsDatePickerTextCtrl *>(this));
    
    // sample string date: it contains more char to have more space
    wxString sep = "/";
    sep = m_separator;
    wxString sdt = wxString(wxT("07")) + sep + wxT("05") + sep + wxT("201000");
    
    int x, y;
    dc.GetTextExtent(sdt, &x, &y);

#ifdef __WXGTK__
	x += 6;		// some more room for GTK
#endif

    wxSize best = wxTextCtrl::DoGetBestSize();
    best.SetWidth( x );
    CacheBestSize(best);
    return best;
}

wxDateTime fsDatePickerTextCtrl::GetValueDT() const
{
    wxString value = GetValue();
    if ( value.IsEmpty() ) {
        return wxInvalidDateTime;
    }
    value = FormatDateText( value, m_separator );
    wxDateTime dt;
    dt.ParseFormat( value, m_format );
    return dt;
}

void fsDatePickerTextCtrl::SetValueDT(const wxDateTime& date)
{
    if ( date.IsValid() )
        SetValue( date.Format(m_format) );
    else
        SetValue( wxEmptyString );
}

void fsDatePickerTextCtrl::OnChar(wxKeyEvent &event)
{
    HandleOnChar( event, m_separator, GetValue() );
}

void fsDatePickerTextCtrl::OnKeyDown(wxKeyEvent &event)
{
    wxDateTime dt = GetValueDT();
    if( HandleOnKeyDown(event, dt) ) {
        SetValueDT( dt );
    }
}

void fsDatePickerTextCtrl::OnKillFocus(wxFocusEvent &event)
{
    wxString value = GetValue();
    if( !value.IsEmpty() ) {
        value = FormatDateText( value, m_separator );
        wxDateTime dt; 
        dt.ParseFormat( value, m_format );
        if ( !dt.IsValid() ) {
            SetValue( wxEmptyString );
        }
        else {
            SetValue( value );
        }
    }

    event.Skip();
}


