/*
	This file is part of vvvP (Virtual Volumes View Photo edition)

	Copyright (C) 2009, the vvvP Development team

	Author: Fulvio Senore

    vvvP is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    vvvP is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    aint with vvvP; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <wx/app.h>
#include <wx/scrolwin.h>
#include <wx/tokenzr.h>
#include <wx/dir.h>
#include <wx/image.h>
#include <wx/bitmap.h>
#include <wx/mstream.h>
#include <wx/file.h>
#include <wx/log.h>
#include <wx/thread.h>

#include "catalog_images_functions.h"
#include "vvvpapp.h"
#include "sqlstrings.h"
#include "db_sequences.h"
#include "sha1wrapper.h"
#include "long_text_question.h"
#include "long_text_info.h"
#include "image_preview.h"
#include "fs_image.h"

#include <exiv2/image.hpp>
#include <exiv2/exif.hpp>

using namespace FSDB;

/////////////////////////////////////////////////////////////////
// thread used to load a file to a buffer
// it will compute the SHA1 hash
/////////////////////////////////////////////////////////////////
class ReadImageThread : public wxThread
{
public:

	ReadImageThread( wxFileName _fn );
	virtual void *Entry();
	wxString GetSHA1() { return sha1; }

private:

	wxString sha1;
	wxFileName fn;
};

ReadImageThread::ReadImageThread( wxFileName _fn ) : wxThread( wxTHREAD_JOINABLE )
{
	this->fn = _fn;
}

void *ReadImageThread::Entry()
{
	wxString fileName = fn.GetFullPath();
	wxFile fin( fileName );
	wxUint8 *fileBuffer = new wxUint8[fn.GetSize().ToULong()];
	fin.Read( fileBuffer, fn.GetSize().ToULong() );
	fin.Close();
	// compute the SHA1 hash
	sha1 = CSHA1Wrapper::ComputeHash( fileBuffer, fn.GetSize().ToULong() );
	delete fileBuffer;

	return NULL;
}



CCatalogImagesFunctions::CCatalogImagesFunctions( wxWindow *_parent, bool *_cancelVariable, wxGauge *_progressBar )
{
	catalogRemovableLocation = false;
	catalogFixedLocation = false;
	cancelVariable = _cancelVariable;
	progressBar = _progressBar;
	parent = _parent;

	CAppSettings& as = wxGetApp().GetAppSettings();
	if( as.m_EnableColorManagement ) {
		// we will convert thumbnails to sRGB color space
		m_ColorProfilesHandler.Assign_sRGBOutputProfile();
	}
}

CCatalogImagesFunctions::~CCatalogImagesFunctions(void)
{
}

void CCatalogImagesFunctions::SetRemovableData( wxString _volumeName, wxString _volumePath, wxString _removablePath ) {
	volumeName = _volumeName;
	volumePath = _volumePath;
	removablePath = _removablePath;
	catalogRemovableLocation = true;

	volumePath = AddTrailingPathSeparator( volumePath );
	removablePath = AddTrailingPathSeparator( removablePath );

	// determine the relative path
	relativeRemovablePath = removablePath.Right( removablePath.Len() - volumePath.Len() );
	if( relativeRemovablePath.StartsWith( wxString(wxFileName::GetPathSeparator()) )) {
		relativeRemovablePath = relativeRemovablePath.Right( relativeRemovablePath.Len() - 1 );
	}
	if( relativeRemovablePath.EndsWith( wxString(wxFileName::GetPathSeparator()) )) {
		relativeRemovablePath = relativeRemovablePath.Left( relativeRemovablePath.Len() - 1 );
	}
}

void CCatalogImagesFunctions::SetFixedData( wxString _fixedLocationName, wxString _fixedLocationPath, wxString _fixedPath ) {
	fixedLocationName = _fixedLocationName;
	fixedLocationPath = _fixedLocationPath;
	fixedPath = _fixedPath;
	catalogFixedLocation = true;

	fixedLocationPath = AddTrailingPathSeparator( fixedLocationPath );
	fixedPath = AddTrailingPathSeparator( fixedPath );

	// determine the relative path
	relativeFixedPath = fixedPath.Right( fixedPath.Len() - fixedLocationPath.Len() );
	if( relativeFixedPath.StartsWith( wxString(wxFileName::GetPathSeparator()) )) {
		relativeFixedPath = relativeFixedPath.Right( relativeFixedPath.Len() - 1 );
	}
	if( relativeFixedPath.EndsWith( wxString(wxFileName::GetPathSeparator()) )) {
		relativeFixedPath = relativeFixedPath.Left( relativeFixedPath.Len() - 1 );
	}
}

wxString CCatalogImagesFunctions::Catalog() {

	wxASSERT( catalogRemovableLocation || catalogFixedLocation );
	
	filesListErrorSHA1.clear();
	filesListErrorSizeDate.clear();
	filesListCorruptedImages.clear();
	
	nUpdatedFiles = nAddedFiles = nDeletedFiles = nDeletedFolders = nUnchangedFiles = 0;

	// primary key of the removable and fixed volume in the VOLUMES table
	int volumeID, fixedLocationID;

	// start a transaction
	Database db = wxGetApp().GetDatabase();
	db->TransactionStart();

	int removableFatherID = -1, fixedFatherID = -1;
	if( catalogRemovableLocation ) {
		volumeID = GetVolumeID( db, volumeName, wxT("R") );
		removableFatherID = GetFolderFatherID( db, relativeRemovablePath, volumeID );
	}
	if( catalogFixedLocation ) {
		fixedLocationID = GetVolumeID( db, fixedLocationName, wxT("F") );
		fixedFatherID = GetFolderFatherID( db, relativeFixedPath, fixedLocationID );
	}

	// catalog both removable and fixed location
	if( catalogRemovableLocation && catalogFixedLocation ) {
		CatalogSingleFolder( db, volumePath, fixedLocationPath, relativeRemovablePath, 
							 relativeFixedPath, volumeID, fixedLocationID, removableFatherID, fixedFatherID ); 
	}
	// catalog only the removable location
	if( catalogRemovableLocation && !catalogFixedLocation ) {
		CatalogSingleFolder( db, volumePath, wxEmptyString, relativeRemovablePath, 
							 wxEmptyString, volumeID, -1, removableFatherID, -1 ); 
	}
	// catalog only the fixed location
	if( !catalogRemovableLocation && catalogFixedLocation ) {
		CatalogSingleFolder( db, fixedLocationPath, wxEmptyString, relativeFixedPath, 
							 wxEmptyString, fixedLocationID, -1, fixedFatherID, -1 ); 
	}

	if( *cancelVariable ) {
		db->TransactionRollback();
		return wxEmptyString;
	}

	if( ReportCatalogResult() ) {
		db->TransactionCommit();
	}
	else {
		db->TransactionRollback();
	}

	return wxEmptyString;
}

int CCatalogImagesFunctions::GetVolumeID( FSDB::Database db, wxString volumeName, wxString volumeType ) {

	int retVal;

	wxASSERT( volumeType == wxT("R") || volumeType == wxT("F") );

	FSDB_CREATE_RS_READWRITE_FWDONLY( db, rs );
	
	// check if the volume is already present in the database
	std::string sql = CSqlStrings::SelectVolumeByName( volumeName, volumeType );
	rs.Open( sql, "VOLUMES", "VOLUME_ID" );
	bool rsEmpty = rs.IsEmpty();
	if( !rsEmpty ) {
		// present, get the ID
		retVal = rs["VOLUME_ID"];
	}
	rs.Close();

	// not present, add it
	if( rsEmpty ) {
		sql = CSqlStrings::SelectVolumesEmpty();
		rs.Open( sql, "VOLUMES", "VOLUME_ID" );
		rs.AddNew();
		rs["VOLUME_NAME"] = CUtils::DBwx2std( volumeName );
		retVal = CDbSequences::GetNextVolumesPK();
		rs["VOLUME_ID"] = retVal;
		rs["VOLUME_TYPE"] = CUtils::DBwx2std( volumeType );
		rs.Update();
		rs.Close();
	}

	return retVal;
}

int CCatalogImagesFunctions::GetSingleFolderID( FSDB::Database db, int volumeID, int fatherID, wxString folderName ) {
	int retVal;

	FSDB_CREATE_RS_READWRITE_FWDONLY( db, rs );

	// check if the folder is already present in the database
	std::string sql = CSqlStrings::SelectPathByName( folderName, volumeID, fatherID );
	rs.Open( sql, "PATHS", "PATH_ID" );
	bool rsEmpty = rs.IsEmpty();
	if( !rsEmpty ) {
		// present, get the ID
		retVal = rs["PATH_ID"];
	}
	rs.Close();

	// not present, add it
	if( rsEmpty ) {
		sql = CSqlStrings::SelectPathsEmpty();
		rs.Open( sql, "PATHS", "PATH_ID" );
		rs.AddNew();
		rs["PATH_NAME"] = CUtils::DBwx2std( folderName );
		retVal = CDbSequences::GetNextPathsPK();
		rs["PATH_ID"] = retVal;
		rs["VOLUME_ID"] = volumeID;
		if( fatherID != -1 )
			rs["FATHER_ID"] = fatherID;
		rs.Update();
		rs.Close();
	}

	return retVal;
}

int CCatalogImagesFunctions::GetFolderID( FSDB::Database db, wxString path, int volumeID ) {

	int retVal = -1;

	// first get the root folder
	retVal = GetSingleFolderID( db, volumeID, -1, wxEmptyString );

	// split path in the sequence of its folders
	if( !path.empty() ) {
		wxStringTokenizer tkz( path, wxFileName::GetPathSeparator() );
		while( tkz.HasMoreTokens() ) {
			wxString folder = tkz.GetNextToken();
			retVal = GetSingleFolderID( db, volumeID, retVal, folder );
		}
	}

	return retVal;
}

int CCatalogImagesFunctions::GetFolderFatherID( FSDB::Database db, wxString path, int volumeID ) {

	if( path.empty() )
		return -1;

	wxChar sep = wxFileName::GetPathSeparator();
	wxString s = path.BeforeLast( sep );
	return GetFolderID( db, s, volumeID );
}

void CCatalogImagesFunctions::CatalogSingleFolder( FSDB::Database db, 
		                      wxString mainLocation, 
							  wxString secondaryLocation, 
							  wxString mainPath, 
							  wxString secondaryPath, 
							  int mainVolumeID, 
							  int secondaryVolumeID, 
							  int mainFatherID,
							  int secondaryFatherID) {

	// used when updating: it will contain the list of files already stored in the database for a certain path
	map<wxString, CFileData> folderFilesMain;		// for the main location
	map<wxString, CFileData> folderFilesSecondary;	// for the secondary location
	map<wxString, CFileData>::iterator ffi;

	// full path to the current folder
	wxString curPathMain = mainLocation + mainPath;

	// name of the current folder
	wxString curFolderName;
	if( mainPath.empty() ) {
		curFolderName = wxEmptyString;
	}
	else {
		curFolderName = mainPath.AfterLast( wxFileName::GetPathSeparator() );
	}

	// get the ID of the current folder, or create it	
	int mainPathID = GetSingleFolderID( db, mainVolumeID, mainFatherID, curFolderName );

	int secondaryPathID;
	wxString curPathSecondary;
	if( !secondaryLocation.empty() ) {
		// we also handle a secondary location
		curPathSecondary = secondaryLocation + secondaryPath;
		secondaryPathID = GetSingleFolderID( db, secondaryVolumeID, secondaryFatherID, curFolderName );
	}
	else {
		secondaryPathID = -1;
	}

	// read from the database the files already cataloged for this folder
	ReadFolderFilesFromDB( folderFilesMain, mainPathID );
	if( secondaryPathID != -1 ) {
		ReadFolderFilesFromDB( folderFilesSecondary, secondaryPathID );
	}

	// create the recordsets needed to add new files
	FSDB_CREATE_RS_READWRITE_FWDONLY( db, rsFiles );
	std::string sql = CSqlStrings::SelectFilesEmpty();
	rsFiles.Open( sql, "FILES", "FILE_ID" );
	FSDB_CREATE_RS_READWRITE_FWDONLY( db, rsThumbnails );
	sql = CSqlStrings::SelectThumbnailsEmpty();
	rsThumbnails.Open( sql, "THUMBNAILS", "FILE_ID_TH" );
	FSDB_CREATE_RS_READWRITE_FWDONLY( db, rsPathsFiles );
	sql = CSqlStrings::SelectPathsFilesEmpty();
	rsPathsFiles.Open( sql, "PATHS_FILES", "PF_ID" );

	// read all file names from the main file system
	wxDir dir(curPathMain);
	wxString fileName;
	bool cont = dir.GetFirst(&fileName, wxT(""), wxDIR_FILES );
	while( cont ) {

		// update the progress bar in the dialog
		progressBar->SetValue( nAddedFiles + nUpdatedFiles + nUnchangedFiles );
		progressBar->Refresh();
		progressBar->Update();
		wxGetApp().Yield();	// here *cancelVariable might become true

		// check if the user asked to interrupt
		if( *cancelVariable ) {
			if( CUtils::MsgAskNo(_("Do you want to stop cataloging?") ) ) {
				return;
			}
			else {
				*cancelVariable = false;
			}
		}

		wxFileName fnPrimary( curPathMain, fileName );
		// see if the file is already present in the main location
		ffi = folderFilesMain.find( fileName );
		if( ffi == folderFilesMain.end() ) {
			// not found in the main location, add the file to the database
			int spID = secondaryPathID;
			wxFileName fnSecondary;
			if( secondaryPathID != -1 ) {
				// see if we find it in the secondary location. if we find it we will not add it for this location
				ffi = folderFilesSecondary.find( fileName );
				if( ffi != folderFilesSecondary.end() ) {
					// found in the secondary, do not add it
					spID = -1;
				}
				else {
					// not found, we should add it to the secondary location
					// check if the file exists
					fnSecondary.Assign( curPathSecondary, fileName );
					if( !fnSecondary.FileExists() ) {
						spID = -1;
					}
					else {
						//// check if size and datetime are the same
						//if( fnPrimary.GetSize() != fnSecondary.GetSize() ||
						//	!fnPrimary.GetModificationTime().IsEqualUpTo(fnSecondary.GetModificationTime(), wxTimeSpan::Seconds(4)) ) {
						// check if size is the same. removed the datetime check because identical images might have diffent datetimes
						// in the main and secondary locations. The SHA1 hash check will verify if files are identical.
						if( fnPrimary.GetSize() != fnSecondary.GetSize() ) {
								spID = -1;
								filesListErrorSizeDate.push_back( fnPrimary.GetFullPath() );
								filesListErrorSizeDate.push_back( fnSecondary.GetFullPath() );
						}
					}
				}
			}
			AddFileRow( db, rsPathsFiles, rsFiles, rsThumbnails, mainPathID, spID, fnPrimary, fnSecondary );
			nAddedFiles++;
		}
		else {
			// found in the main location
			ffi->second.IsStillThere = true;
			// check if file date or size have changed
			wxDateTime fTime = fnPrimary.GetModificationTime();
			wxULongLong fSize = fnPrimary.GetSize();
			if( !fTime.IsEqualUpTo(ffi->second.DateTime, wxTimeSpan::Seconds(10)) || fSize != ffi->second.FileSize ) {
				// here if date and/or size have changed
				// check the hash to see if the file content has changed
				wxString hashPrimary = ComputeFileSHA1( fnPrimary );
				if( hashPrimary != ffi->second.Sha1 ) {
					// file content is different: the simplest thing is to catalog it again
					// we will update the FILES row
					// we will delete all entries of the file in the PATHS_FILES, thumbnails and metadata tables
					//   and we will insert those info again
					int spID = secondaryPathID;
					wxFileName fnSecondary;
					if( secondaryPathID != -1 ) {
						// check if the file exists in the secondary location
						fnSecondary.Assign( curPathSecondary, fileName );
						if( !fnSecondary.FileExists() ) {
							spID = -1;
						}
						else {
							// check if size and datetime are the same
							if( fnPrimary.GetSize() != fnSecondary.GetSize() ) {
									spID = -1;
									filesListErrorSizeDate.push_back( fnPrimary.GetFullPath() );
									filesListErrorSizeDate.push_back( fnSecondary.GetFullPath() );
							}
						}
					}
					UpdateFileRow( db, rsPathsFiles, rsThumbnails, mainPathID, spID, fnPrimary, fnSecondary, ffi->second.FileID );
				}
				else {
					// file content is the same, only date has changed
					UpdateFileDateSize( ffi->second.FileID, fSize, CUtils::dt_wx2fsdb(fTime) );
				}
				nUpdatedFiles++;
			}
			else {
				nUnchangedFiles++;
			}
		}
		cont = dir.GetNext(&fileName);
	}
	rsFiles.Close();
	rsThumbnails.Close();
	rsPathsFiles.Close();

	// scan the map looking for files to be deleted
	ffi = folderFilesMain.begin();
	while( ffi != folderFilesMain.end() ) {
		if( !ffi->second.IsStillThere ) {
			FSDB::Parameters params = ParametersFactory();
			params->Append( static_cast<int32_t>(ffi->second.FileID) );
			std::string sql = CSqlStrings::DeletePhysicalFile();
			db->ExecuteNoReturn( sql, params );
			nDeletedFiles++;
		}
		ffi++;
	}

	// read from the database the subfolder already cataloged for this folder
	ReadFoldersFromDB( folderFilesMain, mainVolumeID, mainPathID );

	// now read all the subfolders
	cont = dir.GetFirst(&fileName, wxT(""), wxDIR_DIRS );
	while( cont ) {
		// see if the folder is already present
		ffi = folderFilesMain.find( fileName );
		if( ffi != folderFilesMain.end() ) {
			ffi->second.IsStillThere = true;
		}
		
		wxString newMainPath;
		if( mainPath.empty() ) {
			newMainPath = fileName;
		}
		else {
			newMainPath = mainPath + wxFileName::GetPathSeparator() + fileName;
		}
		wxString newSecondaryPath;
		if( secondaryPath.empty() ) {
			newSecondaryPath = fileName;
		}
		else {
			newSecondaryPath = secondaryPath + wxFileName::GetPathSeparator() + fileName;
		}

		wxFileName dirName( mainPath, wxEmptyString );
		dirName.AppendDir( fileName );
		CatalogSingleFolder( db, mainLocation, secondaryLocation, newMainPath, newSecondaryPath, mainVolumeID, secondaryVolumeID, mainPathID, secondaryPathID );
		cont = dir.GetNext(&fileName);

		if( *cancelVariable ) return;
	}
							  
	// scan the map looking for folders to be deleted
	ffi = folderFilesMain.begin();
	while( ffi != folderFilesMain.end() ) {
		if( !ffi->second.IsStillThere ) {
			FSDB::Parameters params = ParametersFactory();
			params->Append( static_cast<int32_t>(ffi->second.FileID) );
			std::string sql = CSqlStrings::DeletePhysicalFolder();
			db->ExecuteNoReturn( sql, params );
			nDeletedFolders++;
		}
		ffi++;
	}
}

wxString CCatalogImagesFunctions::AddTrailingPathSeparator( wxString path ) {
	wxString sep(wxFileName::GetPathSeparator());
	if( !path.EndsWith(sep) )
		path += sep;
	return path;
}

void CCatalogImagesFunctions::ReadFolderFilesFromDB( map<wxString, CFileData> &m, int pathID ) {
	m.clear();
	Database db = wxGetApp().GetDatabase();
	FSDB_CREATE_RS_READONLY_FWDONLY( db, rs );
	std::string sql = CSqlStrings::SelectFilesInAPath( pathID );
	rs.Open( sql );
	while( !rs.IsEOF() ) {
		CFileData *fd = new CFileData();
		fd->FileID = rs["FILE_ID"].GetAsInteger();
		fd->DateTime = CUtils::dt_fsdb2wx( rs["FILE_DATETIME"] );
		fd->FileSize = CUtils::i64_fsdb2wx( rs["FILE_SIZE"].GetAsLargeInt() );
		fd->Sha1 = CUtils::DBstd2wx(rs["SHA1_HASH"]);

		wxString fileName = CUtils::DBstd2wx(rs["FILE_NAME"]);
		m.insert( FDPair(fileName, *fd) );
		delete fd;

		rs.MoveNext();
	}
	rs.Close();
}

void CCatalogImagesFunctions::ReadFoldersFromDB( map<wxString, CFileData> &m, int volumeID, int pathID ) {
	m.clear();
	Database db = wxGetApp().GetDatabase();
	FSDB_CREATE_RS_READONLY_FWDONLY( db, rs );
	std::string sql = CSqlStrings::SelectPathsList( volumeID, pathID );
	rs.Open( sql );
	while( !rs.IsEOF() ) {
		// we use the CFileData structure to store only the path ID
		CFileData *fd = new CFileData();
		fd->FileID = rs["PATH_ID"].GetAsLargeInt();

		wxString fileName = CUtils::DBstd2wx(rs["PATH_NAME"]);
		m.insert( FDPair(fileName, *fd) );
		delete fd;

		rs.MoveNext();
	}
	rs.Close();
}
	
void CCatalogImagesFunctions::StoreFileRow( FSDB::Database db, 
										FSDB::Recordset rsPathsFiles, 
										FSDB::Recordset rsFiles, 
										FSDB::Recordset rsThumbnails, 
										int mainPathID, 
										int secondaryPathID, 
										wxFileName fnPrimary,
										wxFileName fnSecondary,
										int fileID ) {

	wxULongLong fileSize = fnPrimary.GetSize();
	FSDB::Timestamp fileDateTime = CUtils::dt_wx2fsdb( fnPrimary.GetModificationTime() );
	wxString sha1Secondary;
	wxImage image = wxNullImage;	// the image stored in the file
	CMemoryChunk colorProfile;

	ReadImageThread *secThread = NULL;

	if( secondaryPathID != -1 ) {
		// start a thread to read the file from the secondary location
		// if we read both files from slow devices using a thread should make things much faster
		secThread = new ReadImageThread( fnSecondary );
		if( secThread->Create() != wxTHREAD_NO_ERROR ) {
			CUtils::MsgErr( _("Unable to create thread") );
		}
		secThread->Run();
	}

	// read the primary file in a memory buffer
	wxString fileName = fnPrimary.GetFullPath();
	wxFile fin( fileName );
	wxUint8 *fileBuffer = new wxUint8[fileSize.ToULong()];
	fin.Read( fileBuffer, fileSize.ToULong() );
	fin.Close();
	// compute the SHA1 hash of the primary file
	wxString sha1Primary = CSHA1Wrapper::ComputeHash( fileBuffer, fileSize.ToULong() );

	if( secondaryPathID != -1 ) {
		// wait for the thread
		secThread->Wait();
		sha1Secondary = secThread->GetSHA1();
		delete secThread;
		// check if the files are identical
		if( sha1Primary != sha1Secondary ) {
			filesListErrorSHA1.push_back( fnPrimary.GetFullPath() );
			filesListErrorSHA1.push_back( fnSecondary.GetFullPath() );
			secondaryPathID = -1;	// do not catalog the file in the secondary location
		}
	}

	// read basic metadata
	CUtils::ImageRotation imageRotation;
	wxDateTime imageDateTime;
	int imageWidth = -1, imageHeight = -1;
	CMetadata::ColorSpace exifColorSpace;
	CMetadata::ReadBasicMetadata( fileBuffer, fileSize.ToULong(), imageRotation, imageDateTime, imageWidth, imageHeight, exifColorSpace );
	// write the FILES row
	rsFiles["FILE_ID"] = fileID;
	rsFiles["FILE_SIZE"] = CUtils::i64_wx2fsdb( fileSize );
	rsFiles["FILE_NAME"] = CUtils::DBwx2std( fnPrimary.GetFullName() );
	wxString ext = fnPrimary.GetExt();
	if( ext.Len() > 30 ) ext = wxEmptyString;	// such a long extension is surely a meaningless temporary file
	rsFiles["FILE_EXT"] = CUtils::DBwx2std( ext );
	rsFiles["FILE_DATETIME"] = fileDateTime;
	if( imageDateTime.IsValid() ) {
		rsFiles["IMAGE_DATETIME"] = CUtils::dt_wx2fsdb( imageDateTime );
	}
	else {
		rsFiles["IMAGE_DATETIME"] = fileDateTime;	// set to the file datetime since there is no exif datetime
	}
	rsFiles["SHA1_HASH"] = CUtils::DBwx2std( sha1Primary );
	rsFiles["ROTATION"] = static_cast<int>( imageRotation );
	
	// first try to extract the thumbnail from embedded previews
	CAppSettings& as = wxGetApp().GetAppSettings();
	bool tryPreview = true;
	if( !as.m_UseJpegPreview && IsHandledImageType(ext) ) tryPreview = false;
	CImagePreview::ReturnValue r;
	if( tryPreview )
		r = CImagePreview::ExtractImagePreview( fileName, as.m_ThumbnailSize, image, colorProfile, exifColorSpace );
	else
		r = CImagePreview::rtNoPreviews;	// just to load the full image
	if( r != CImagePreview::rtOk ) {
		// unable to use prevew images, load the whole image
		if( IsHandledImageType(ext) ) {
			// convert the buffer to an image
			wxMemoryInputStream mis( fileBuffer, fileSize.ToULong() );
			{
				wxLogNull ln;
				fsImage img( mis );
				if( img.IsOk() ) {
					colorProfile.SetData( img.GetICCProfileAddress(), img.GetICCProfileSize() );
					img.DeleteICCProfile();
				}
				image = img;	// assign to a variable with a larger scope
			}
			if( !image.IsOk() ) {
				filesListCorruptedImages.push_back( fileName );
			}
		}
	}
	if( image.IsOk() ) {
		// resize to thumbnail size
		int w, h;
		image = ResizeImageToThumbnailSize( image, as.m_ThumbnailSize, w, h );

		// convert to sRGB
		CAppSettings& as = wxGetApp().GetAppSettings();
		if( as.m_EnableColorManagement ) {
			if( colorProfile.GetAddress() == NULL && exifColorSpace == CMetadata::csAdobeRGB ) {
				// the image does not contain a profile but Exif data say that it uses Adobe RGB
				m_ColorProfilesHandler.TransformImageFromAdobeRGB( image );
			}
			else {
				m_ColorProfilesHandler.TransformImage( image, colorProfile );
			}
		}
	}

	// store the image size in the database
	if( image.IsOk() ) {
		rsFiles["WIDTH"] = image.GetWidth();
		rsFiles["HEIGHT"] = image.GetHeight();
	}
	else {
		rsFiles["WIDTH"].SetNull();
		rsFiles["HEIGHT"].SetNull();
	}

	rsFiles.Update();

	// add the PATHS_FILES row
	rsPathsFiles.AddNew();
	rsPathsFiles["FILE_ID_PF"] = fileID;
	rsPathsFiles["PATH_ID_PF"] = mainPathID;
	rsPathsFiles.Update();

	if( secondaryPathID != -1 ) {
		// add the secondary path
		rsPathsFiles.AddNew();
		rsPathsFiles["FILE_ID_PF"] = fileID;
		rsPathsFiles["PATH_ID_PF"] = secondaryPathID;
		rsPathsFiles.Update();
	}

	// add the THUMBNAILS row
	if( image.IsOk() ) {
		AddThumbnailRow( rsThumbnails, image, fileID );
	}

	// store image metadata
	CMetadata::ReadFileTags( fileName, fileID, db );

	delete fileBuffer;
}

void CCatalogImagesFunctions::AddFileRow( FSDB::Database db, 
										FSDB::Recordset rsPathsFiles, 
										FSDB::Recordset rsFiles, 
										FSDB::Recordset rsThumbnails, 
										int mainPathID, 
										int secondaryPathID, 
										wxFileName fnPrimary,
										wxFileName fnSecondary ) {
											
	// add the FILES row
	int fileID = CDbSequences::GetNextFilesPK();
	rsFiles.AddNew();

	StoreFileRow( db, rsPathsFiles, rsFiles, rsThumbnails, mainPathID, secondaryPathID, fnPrimary, fnSecondary, fileID );
}

void CCatalogImagesFunctions::UpdateFileRow( FSDB::Database db, 
										FSDB::Recordset rsPathsFiles, 
										FSDB::Recordset rsThumbnails, 
										int mainPathID, 
										int secondaryPathID, 
										wxFileName fnPrimary,
										wxFileName fnSecondary,
										int fileID ) {
											
	FSDB_CREATE_RS_READWRITE_FWDONLY( db, rsFiles );
	std::string sql = CSqlStrings::SelectFileFromID( fileID );
	rsFiles.Open( sql, "FILES", "FILE_ID" );
	if( rsFiles.IsEmpty() ) {
		// not found, add a new row
		AddFileRow( db, rsPathsFiles, rsFiles, rsThumbnails, mainPathID, secondaryPathID, fnPrimary, fnSecondary );
	}
	else {
		// delete thumbnails, metadata and PATH_FILES entries that will be recreated by StoreFileRow()
		std::string sql = CSqlStrings::DeleteFileThumbnail( fileID );
		db->ExecuteNoReturn( sql );
		CMetadata::DeleteFileTags( fileID, db );
		sql = CSqlStrings::DeleteFilePathsFiles( fileID );
		db->ExecuteNoReturn( sql );

		rsFiles.Edit();
		StoreFileRow( db, rsPathsFiles, rsFiles, rsThumbnails, mainPathID, secondaryPathID, fnPrimary, fnSecondary, fileID );
	}
	rsFiles.Close();
}

void CCatalogImagesFunctions::StoreThumbnailRow( FSDB::Recordset rsThumbnails, 
		                  wxImage image,
						  int fileID ) {

	// create the thumbnail blob
	int w, h;
	char *buf;
	int bufSize;
	ConvertImageToBlob( image, buf, bufSize );
	w = image.GetWidth();
	h = image.GetHeight();

	// save in the database
	rsThumbnails["FILE_ID_TH"] = fileID;
	rsThumbnails["THUMBNAIL"].SetAsBlobNoMemoryAllocation( buf, bufSize );
	rsThumbnails["TH_WIDTH"] = w;
	rsThumbnails["TH_HEIGHT"] = h;

}

void CCatalogImagesFunctions::AddThumbnailRow( FSDB::Recordset rsThumbnails, 
		                  wxImage image,
						  int fileID ) {

	rsThumbnails.AddNew();
	StoreThumbnailRow( rsThumbnails, image, fileID );
	rsThumbnails.Update();
}

void CCatalogImagesFunctions::UpdateThumbnailRow( wxImage image,
						  int fileID ) {

	Database db = wxGetApp().GetDatabase();
	FSDB_CREATE_RS_READWRITE_FWDONLY( db, rs );
	std::string sql = CSqlStrings::SelectFileThumbnail( fileID );
	rs.Open( sql, "THUMBNAILS", "FILE_ID_TH" );
	if( rs.IsEmpty() ) {
		// no thumbnail, add a new one
		AddThumbnailRow( rs, image, fileID );
	}
	else {
		rs.Edit();
		StoreThumbnailRow( rs, image, fileID );
		rs.Update();
	}
	rs.Close();
}

wxImage CCatalogImagesFunctions::ResizeImageToThumbnailSize( const wxImage &image, const int thSize, int &thWidth, int &thHeight )
{
	// compute the thumbnail size
	thWidth = image.GetWidth();
	thHeight = image.GetHeight();
	if( thWidth > thHeight ) {
		thHeight = thHeight * (((float)thSize)/thWidth);
		thWidth = thSize;
	}
	else {
		thWidth = thWidth * (((float)thSize)/thHeight);
		thHeight = thSize;
	}
	// resize the image
	const fsImage *ip = static_cast<const fsImage*>( &image );
	wxImage imgResized = ip->ScaleFS( thWidth, thHeight );

	return imgResized;
}

void CCatalogImagesFunctions::ConvertImageToBlob( const wxImage &image, char* &buf, int &bufSize )
{
	wxImage imgResized = image;
	imgResized.SetOption( wxIMAGE_OPTION_QUALITY, 90 );
	wxMemoryOutputStream mos;
	imgResized.SaveFile( mos, wxBITMAP_TYPE_JPEG );
	// set return values
	bufSize =  (int) mos.GetSize();
	buf = new char[bufSize];
	mos.CopyTo( buf, bufSize );
}

void CCatalogImagesFunctions::UpdateFileDateSize( int fileID, const wxULongLong &fileSize, const FSDB::Timestamp &fileDateTime ) {
	Database db = wxGetApp().GetDatabase();
	FSDB_CREATE_RS_READWRITE_FWDONLY( db, rs );
	std::string sql = CSqlStrings::SelectFileFromID( fileID );
	rs.Open( sql, "FILES", "FILE_ID" );
	rs.Edit();
	rs["FILE_DATETIME"] = fileDateTime;
	rs["FILE_SIZE"] = CUtils::i64_wx2fsdb( fileSize );
	rs.Update();
	rs.Close();
}

bool CCatalogImagesFunctions::ReportCatalogResult() {

	wxString newLine = wxT("\n");

	if( !filesListCorruptedImages.empty() ) {
		CLongTextInfo lti( parent, ID_DIALOG_LONG_TEXT_INFO, _("Corrupted files warning") );
		lti.SetUpperText( _("Unable to load one or more images, they are probably corrupted.\nHere is the list of the files:") );
		wxString filesList = wxEmptyString;
		for( size_t k = 0; k < filesListCorruptedImages.size(); k++ )
			filesList += filesListCorruptedImages[k] + newLine;
		lti.SetLongText( filesList );
		lti.ShowModal();
	}

	wxString stdMsg;
	if( nAddedFiles > 0 ) stdMsg = stdMsg + _("Number of added files:") + wxT(" ") + CUtils::long2string(nAddedFiles) + wxT("\n");
	if( nUpdatedFiles > 0 ) stdMsg = stdMsg + _("Number of updated files:") + wxT(" ") + CUtils::long2string(nUpdatedFiles) + wxT("\n");
	if( nDeletedFiles > 0 ) stdMsg = stdMsg + _("Number of deleted files (not counting those in deleted folders):") + wxT(" ") + CUtils::long2string(nDeletedFiles) + wxT("\n");
	if( nDeletedFolders > 0 ) stdMsg = stdMsg + _("Number of deleted folders:") + wxT(" ") + CUtils::long2string(nDeletedFolders) + wxT("\n");
	if( nUnchangedFiles > 0 ) stdMsg = stdMsg + _("Number of unchanged files:") + wxT(" ") + CUtils::long2string(nUnchangedFiles) + wxT("\n");


	if( filesListErrorSHA1.empty() && filesListErrorSizeDate.empty() ) {
		// no errors, just show the result
		if( nAddedFiles == 0 && nUpdatedFiles == 0 && nDeletedFiles == 0 && nDeletedFolders == 0 ) {
			// nothing has changed, no need to ask
			CUtils::MsgInfo( stdMsg );
			return true;
		}
		else
			return CUtils::MsgAskYes( stdMsg + newLine + _("Is it OK to store these changes in the catalog?") );
	}

	wxString midText = stdMsg + newLine;
	if( !filesListErrorSizeDate.empty() ) {
		midText = midText + _("The following files have a different size, the files in the fixed location have been ignored:") + newLine + newLine;
		bool jump = false;
		for( size_t k = 0; k < filesListErrorSizeDate.size(); k++ ) {
			midText += filesListErrorSizeDate[k] + newLine;
			if( jump ) midText += newLine;	// separe the file couples
			jump = !jump;
		}
		midText += newLine + newLine;
	}
	if( !filesListErrorSHA1.empty() ) {
		midText = midText + _("The following files contain different data, the files in the fixed location have been ignored:") + newLine + newLine;
		bool jump = false;
		for( size_t k = 0; k < filesListErrorSHA1.size(); k++ ) {
			midText += filesListErrorSHA1[k] + newLine;
			if( jump ) midText += newLine;	// separe the file couples
			jump = !jump;
		}
		midText += newLine + newLine;
	}

	CLongTextQuestion dialog( parent, ID_DIALOG_LONG_TEXT_QUESTION, _("Different files warning") );
	dialog.SetUpperText( _("You have chosen to catalog both a fixed and a removable location,\nbut some files with the same name are not identical.\nHere is the list of different files.") );
	dialog.SetMidText( midText );
	dialog.SetLowerText( _("Click Yes to store these changes in the catalog, click No to cancel this task.\nDo you want to continue?") );
	dialog.SetHelpPage( wxT("ltq_catimg_different files.htm") );

	int r = dialog.ShowModal();
	return (r == wxID_YES);
}

wxString CCatalogImagesFunctions::ComputeFileSHA1( const wxFileName &fn ) {
	wxFile fin( fn.GetFullPath() );
	wxULongLong fileSize = fn.GetSize();
	wxUint8 *fileBuffer = new wxUint8[fileSize.ToULong()];
	fin.Read( fileBuffer, fileSize.ToULong() );
	fin.Close();
	wxString sha1 = CSHA1Wrapper::ComputeHash( fileBuffer, fileSize.ToULong() );
	delete fileBuffer;
	return sha1;
}

bool CCatalogImagesFunctions::IsHandledImageType( const wxString &ext )
{
	bool retVal = false;

	if( ext.Upper() == wxT("JPG") ) retVal = true;
	if( ext.Upper() == wxT("JPEG") ) retVal = true;
	if( ext.Upper() == wxT("TIF") ) retVal = true;
	if( ext.Upper() == wxT("PNG") ) retVal = true;

	return retVal;
}

