/*
	This file is part of vvvP (Virtual Volumes View Photo edition)

	Copyright (C) 2008, the vvvP Development team

	Author: Fulvio Senore

    vvvP is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    vvvP is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    aint with vvvP; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include "compare_folders.h"

#include <wx/filename.h>
#include <wx/dir.h>

CCompareFolders::CCompareFolders( wxString _path1, wxString _path2 ) {
	pathPrimary = _path1;
	pathSecondary = _path2;

	wxString sep(wxFileName::GetPathSeparator());

	// make sure that the paths end with a path separator
	if( !pathPrimary.EndsWith(sep) )
		pathPrimary += sep;
	if( !pathSecondary.EndsWith(sep) )
		pathSecondary += sep;
}

CCompareFolders::~CCompareFolders(void)
{
}

void CCompareFolders::ReadAllFiles( wxString basePath, wxString path, std::map<wxString, CFileData> &files, bool recurse ) {

	wxString fileName;

	// builds the physical path for this folder
	wxString physPath = basePath + path;

	// read all the files in the current folder
	wxDir dir(physPath);
	bool cont = dir.GetFirst(&fileName, wxT(""), wxDIR_FILES );
	while( cont ) {
		wxString name;
		if( !path.empty() )
			name = path + wxFileName::GetPathSeparator();
		name += fileName;

		CFileData *fd = new CFileData();
		wxFileName fn( physPath, fileName );
		fd->DateTime = fn.GetModificationTime();
		fd->FileSize = fn.GetSize();

		files.insert( FDPair(name, *fd) );
		delete fd;

		cont = dir.GetNext(&fileName);
	}

	if( !recurse ) return;

	// recurse in the subfolders
	cont = dir.GetFirst(&fileName, wxT(""), wxDIR_DIRS );
	while( cont ) {
		wxString newPath;
		if( !path.empty() )
			newPath = path + wxFileName::GetPathSeparator();
		newPath += fileName;

		ReadAllFiles( basePath, newPath, files, recurse );

		cont = dir.GetNext(&fileName);
	}
}

void CCompareFolders::Compare() {

	// clear the lists
	files1.clear();
	files2.clear();
	filesOnlyPath1.clear();
	filesOnlyPath2.clear();
	filesDifferent.clear();

#ifdef __WXMSW__
	// read all the files in the primary path
	ReadAllFilesWindows( pathPrimary, wxEmptyString, files1, true );
	// read all the files in the secondary path
#else
	ReadAllFiles( pathPrimary, wxEmptyString, files1, true );
#endif

	if( pathSecondary.empty() ) {
		// we only read the files in order to know thier number
		return;
	}

	// read all the files in the secondary path
	ReadAllFilesInSecondaryFolder( pathSecondary, wxEmptyString, pathPrimary, files2 );

	// now check for differences
	std::map<wxString, CFileData>::iterator fiScan;
	std::map<wxString, CFileData>::iterator fiSearch;
	for( fiScan = files2.begin(); fiScan != files2.end(); fiScan++ ) {
		// look for the file in the other folder
		fiSearch = files1.find( fiScan->first );
		if( fiSearch == files1.end() ) {
			// not found in path1
			filesOnlyPath2.push_back( fiScan->first );
		}
		else {
			// remember that we found this file
			fiSearch->second.FoundInTheOtherFolder = true;
			// check if files are different
			if( !fiSearch->second.DateTime.IsEqualUpTo(fiScan->second.DateTime, wxTimeSpan::Seconds(4)) ||
				fiSearch->second.FileSize != fiScan->second.FileSize ) {
					// files are different
					filesDifferent.push_back( fiScan->first );
				}
		}
	}

	// now look for files missing in path2
	for( fiScan = files1.begin(); fiScan != files1.end(); fiScan++ ) {
		if( !fiScan->second.FoundInTheOtherFolder ) {
			filesOnlyPath1.push_back( fiScan->first );
		}
	}
}

void CCompareFolders::ReadAllFilesInSecondaryFolder( wxString basePath, wxString path, wxString primaryPath, 
								   std::map<wxString, CFileData> &files ) {

	// first read all the files in the secondary path, excluding subfolders
	ReadAllFiles( basePath, path, files, false );

	// now analyze all the subfolders contained in the primary path
	wxString fileName;
	wxString physPath = basePath + path;
	wxDir dir(primaryPath);
	bool cont = dir.GetFirst(&fileName, wxT(""), wxDIR_DIRS );
	while( cont ) {
		wxString newPath;
		if( !path.empty() )
			newPath = path + wxFileName::GetPathSeparator();
		newPath += fileName;

		wxString newPhysPath = basePath + newPath;

		if( wxFileName::DirExists(newPhysPath) ) {
			ReadAllFiles( basePath, newPath, files, true );
		}

		cont = dir.GetNext(&fileName);
	}
}

