/*
	This file is part of vvvP (Virtual Volumes View Photo edition)

	Copyright (C) 2008, the vvvP Development team

	Author: Fulvio Senore

    vvvP is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    vvvP is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with vvvP; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

/*
This class implements an array-like opject that holds the data of the files read from the database
*/

#ifndef __FILEDATA_ARRAY_H__
#define __FILEDATA_ARRAY_H__

#include "fsdb.h"
#include "metadata.h"

#include <deque>

#include <wx/string.h>
#include <wx/datetime.h>
#include <wx/longlong.h>
#include <wx/image.h>

class CFileData {

	friend class CFileDataArray;

private:

	int ID;				// primary key of the file in the FILES table (physical view)
	int virtualID;		// primary key in the VIRTUAL_FILES table (virtual view) or -1
	wxString name;
	wxULongLong size;
	int width;			// image width or -1 if NULL
	int height;			// image height or -1 if NULL
	wxString ext;
	wxString description;
	wxDateTime fileDateTime;
	wxDateTime imageDateTime;	// taken form exif if available, otherwise == fileDateTime
	bool _hasThumbnail;		// true if this file has a thumbnail
	wxImage thumbnail;
	int thWidth;		// thumbnail width
	int thHeight;		// thumbnail height
	CUtils::ImageRotation rotation;

	// metadata for this file
	CMetadata metadata;

public:

	int GetID() { return ID; }
	int GetVirtualID() { return virtualID; }
	const wxString& GetName() { return name; }
	const wxULongLong& GetSize() { return size; }
	const int GetWidth() { return width; }
	const int GetHeight() { return height; }
	const wxString& GetExt() { return ext; }
	const wxString& GetDescription() { return description; }
	const wxDateTime& GetFileDateTime() { return fileDateTime; }
	const wxDateTime& GetImageDateTime() { return imageDateTime; }
	const wxImage& GetThumbnailImage() { return thumbnail; }
	const CMetadata& GetMetadata() { return metadata; }
	int GetThumbnailWidth() { return thWidth; }
	int GetThumbnailHeight() { return thHeight; }
	bool HasThumbnail() { return _hasThumbnail; }
	void DeleteThumbnail();		// delete the thumbnail, if any
	CUtils::ImageRotation GetRotation() { return rotation; }
	void SetRotation( CUtils::ImageRotation rot ) { rotation = rot; }
	void SetDescription( wxString descr ) { description = descr; }
	void SetThumbnail( wxImage tmbImage );

	bool operator <( const CFileData &other );

};


class CFileDataArray
{
public:
	CFileDataArray(void);
	~CFileDataArray(void);

	// remove everything from the array
	void Clear();

	// fill the array with data read from a recordset
	// if isVirtual == true data comes from a virtual folder
	// if append == false do not clear the list before adding the new data
	void FillFromRecordset( FSDB::Database db, FSDB::Recordset rs, bool isVirtual, bool append = false );

	// fill the array with data read from a recordset (data are appended)
	// if isVirtual == true data comes from a virtual folder
	// read no more than maxRows rows from the recordset (less if EOF found)
	// return the number of rows that have been read
	int FillFromRecordsetPartial( FSDB::Database db, FSDB::Recordset rs, bool isVirtual, int maxRows );

	// return an element of the array
	CFileData& operator[]( int index ) { return *dataArray[index]; }

	// return the number of elements in the array
	int size() { return (int) dataArray.size(); }

	// sort data
	void Sort();

private:

	/// Data structure that stores the database data
	std::deque<CFileData*> dataArray;

};



#endif
