/*
	This file is part of vvvP (Virtual Volumes View Photo edition)

	Copyright (C) 2009, the vvvP Development team

	Author: Fulvio Senore

    vvvP is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    vvvP is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with vvvP; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

/////////////////////////////////////////////////////////////////////////////
// Name:        fs_imageviewer.h
// Purpose:     A control to show images
// Author:      Fulvio Senore
// Modified by: 
// Created:     04/08/2009 21:08:30
// RCS-ID:      
// Copyright:   
// Licence:     wxWidgets licence
/////////////////////////////////////////////////////////////////////////////

#ifndef _FS_IMAGEVIEWER_H_
#define _FS_IMAGEVIEWER_H_


/*!
 * Includes
 */

////@begin includes
////@end includes

#include "fs_lcms.h"

/*!
 * Forward declarations
 */

////@begin forward declarations
class FS_ImageViewer;
////@end forward declarations

// declaration of custom events for this control
class fsImageViewerEvent : public wxNotifyEvent
{
public:
	fsImageViewerEvent( wxEventType commandType = wxEVT_NULL, int id = 0 ) :
	  wxNotifyEvent( commandType, id ) {}

	fsImageViewerEvent( const fsImageViewerEvent& event ) :
	  wxNotifyEvent( event ) {}

	virtual wxEvent *Clone() const { return new fsImageViewerEvent( *this ); }

	DECLARE_DYNAMIC_CLASS( fsImageViewerEvent );
};

typedef void (wxEvtHandler::*fsImageViewerEventFunction) (fsImageViewerEvent&);

BEGIN_DECLARE_EVENT_TYPES()
	DECLARE_EVENT_TYPE( fsEVT_COMMAND_ZOOM_CHANGED, -1 )
END_DECLARE_EVENT_TYPES()

#define EVT_COMMAND_ZOOM_CHANGED( id, fn) \
	DECLARE_EVENT_TABLE_ENTRY( \
		fsEVT_COMMAND_ZOOM_CHANGED, id, wxID_ANY, \
        (wxObjectEventFunction)(wxEventFunction) wxStaticCastEvent( fsImageViewerEventFunction, &fn ), \
        (wxObject *) NULL \
    ),



/*!
 * FS_ImageViewer class declaration
 */

class FS_ImageViewer: public wxScrolledWindow
{    
    DECLARE_DYNAMIC_CLASS( FS_ImageViewer )
    DECLARE_EVENT_TABLE()

public:
    /// Constructors
    FS_ImageViewer();
    FS_ImageViewer(wxWindow* parent, wxWindowID id = wxID_ANY, const wxPoint& pos = wxDefaultPosition, const wxSize& size = wxSize(100, 100), long style = wxSUNKEN_BORDER|wxHSCROLL|wxVSCROLL);

    /// Creation
    bool Create(wxWindow* parent, wxWindowID id = wxID_ANY, const wxPoint& pos = wxDefaultPosition, const wxSize& size = wxSize(100, 100), long style = wxSUNKEN_BORDER|wxHSCROLL|wxVSCROLL);

    /// Destructor
    ~FS_ImageViewer();

    /// Initialises member variables
    void Init();

    /// Creates the controls and sizers
    void CreateControls();

////@begin FS_ImageViewer event handler declarations

    /// wxEVT_PAINT event handler for ID_FSVIEWER_WINDOW
    void OnPaint( wxPaintEvent& event );

    /// wxEVT_ERASE_BACKGROUND event handler for ID_FSVIEWER_WINDOW
    void OnEraseBackground( wxEraseEvent& event );

    /// wxEVT_LEFT_DOWN event handler for ID_FSVIEWER_WINDOW
    void OnLeftDown( wxMouseEvent& event );

    /// wxEVT_LEFT_UP event handler for ID_FSVIEWER_WINDOW
    void OnLeftUp( wxMouseEvent& event );

    /// wxEVT_MOTION event handler for ID_FSVIEWER_WINDOW
    void OnMotion( wxMouseEvent& event );

    /// wxEVT_MOUSEWHEEL event handler for ID_FSVIEWER_WINDOW
    void OnMouseWheel( wxMouseEvent& event );

////@end FS_ImageViewer event handler declarations

	void OnSize( wxSizeEvent& event );

    void OnChar( wxKeyEvent& event );
    void OnKeyDown(wxKeyEvent& event);
    void OnKeyUp(wxKeyEvent& event);

////@begin FS_ImageViewer member function declarations

    /// Retrieves bitmap resources
    wxBitmap GetBitmapResource( const wxString& name );

    /// Retrieves icon resources
    wxIcon GetIconResource( const wxString& name );
////@end FS_ImageViewer member function declarations

	// set the image to show
	void SetImage( wxImage img );
	// get the current image
	wxImage GetImage() { return m_Image; }

	// set the zoom level
	void SetZoomLevel( float zoom );
	// get the current zoom level
	float GetZoomLevel() { return m_ZoomLevel; }

	// resize the image so that if exactly fills the client area in the horizontal direction
	void FitHorizontal();
	// resize the image so that if exactly fills the client area in the vertical direction
	void FitVertical();
	// resize the image so that it exactly fits the client area, 
	// i.e. the larger possible size that does not require scrollbars
	void Fit();

	// zoom into the image
	void ZoomIn();
	// zoom out of the image
	void ZoomOut();

	// set AutoFit
	void SetAutoFit( bool doFit ) { m_AutoFit = doFit; }
	// get it
	bool GetAutoFit() { return m_AutoFit; }

	// show a text message
	void ShowMessage( wxString msg );

    /// Should we show tooltips?
    static bool ShowToolTips();

////@begin FS_ImageViewer member variables
////@end FS_ImageViewer member variables

protected:

	// available output modes: draw image or write text
	enum OutputMode { omImage, omText };

	// create and show the scaled bitmap to the window
	// if noScrollbars == true the function will make sure that scrollbars are not visible
	//   (used with Fit() because a probable bug caused the scrollbars to show if the window had almost the same proportions than the image)
	// if verticalScrollbar == true the window will have to show a vertical scrollbar, so leave the space for it
	// same for horizontalScrollbar
	void CreateScaledBitmap( bool noScrollbars = false, bool verticalScrollbar = false, bool horizontalScrollbar = false );

	// return true if the whole image is visible (i.e. the window is large enough and there are no scroll bars)
	bool IsWholeImageVisible();

	// return true if the image does not fill completely the client area so some background is visible
	bool IsBackgroundVisible();

	// compute the zoom level for an horizontal fit
	// zoom is the computed zoom level
	// verticalScrollbar is true if the image must show a vertical scrollbar
	void ComputeZoomLevelForHorizontalFit( float &zoom, bool &verticalScrollbar );
	// compute the zoom level for a vertical fit
	void ComputeZoomLevelForVerticalFit( float &zoom, bool &horizontalScrollbar );

	// post a "zoom changed" event to the parent windows
	void PostZoomChangedEvent();

	// the image to show
	wxImage m_Image;

	// bitmap of the scaled image to the desired zoom factor
	wxBitmap m_ScaledBitmap;

	// true if m_ScaledBitmap contains an image different from m_Image
	// it happens when the image has just been changed
	bool m_ImageAndBitmapOutOfSync;

	// current zoom level in % values: 100 means 1:1
	float m_ZoomLevel;

	// increment value used to change the zoom value when zooming in and out
	float m_ZoomDelta;

	// min zoom level
	float m_MinZoomLevel;
	// max zoom level
	float m_MaxZoomLevel;

	// last mouse coordinates, used for mouse scrolling
	wxPoint m_LastMouseCoords;

	// width of the scrollbars, used by Fitxxx() functions
	int m_wScrollbarH, m_wScrollbarV;

	// cursor used when scrolling with the mouse
	wxCursor *m_HandCursor;

	// true if the last command received by the control was "Fit" (used to automatically fit after a resize)
	bool m_LastCommandWasFit;

	// flag for the AutoFit feature: automatically Fit the image when the window is resized if the last command was Fit()
	bool m_AutoFit;

	// current output mode
	OutputMode m_OutputMode;

	// message text that will be written in the control instead of an image
	wxString m_OutputMessage;
};

#endif
    // _FS_IMAGEVIEWER_H_
