/*
	This file is part of vvvP (Virtual Volumes View Photo edition)

	Copyright (C) 2011, the vvvP Development team

	Author: Fulvio Senore

    vvvP is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    vvvP is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with vvvP; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include "wx/wxprec.h"

#ifdef __BORLANDC__
#pragma hdrstop
#endif

#ifndef WX_PRECOMP
#include "wx/wx.h"
#endif

#include "fs_lcms.h"


// dummy function used to force linking to sqrtf
// used to avoid unresolved symbol error from linker
//extern "C" {
//float sqrtf( float n ) {
//	return (float) sqrt( n );
//}
//}

CColorProfile::CColorProfile(void)
{
	m_OutputProfile = NULL;
	m_OutputProfileIs_sRGB = false;

	m_AdobeRGBProfile = Create_AdobeRGB();
}

CColorProfile::~CColorProfile(void)
{
	if( m_OutputProfile != NULL ) {
		cmsCloseProfile( m_OutputProfile );
	}
	if( m_AdobeRGBProfile != NULL ) {
		cmsCloseProfile( m_AdobeRGBProfile );
	}
}

void CColorProfile::Assign_sRGBOutputProfile()
{
	if( m_OutputProfile != NULL ) {
		cmsCloseProfile( m_OutputProfile );
	}
	m_OutputProfile = cmsCreate_sRGBProfile();
	m_OutputProfileIs_sRGB = true;
}

CColorProfile::ReturnValue CColorProfile::AssignFileOutputProfile( const wxString& profileFile )
{
	if( m_OutputProfile != NULL ) {
		cmsCloseProfile( m_OutputProfile );
	}
	
	if( !wxFileExists(profileFile) ) {
		Assign_sRGBOutputProfile();
		m_OutputProfileIs_sRGB = true;
		return rvFileMissing;
	}
	
	m_OutputProfile = cmsOpenProfileFromFile( profileFile.char_str(), "r" );

	if( m_OutputProfile == NULL ) {
		Assign_sRGBOutputProfile();
		m_OutputProfileIs_sRGB = true;
		return rvFileInvalid;
	}

	m_OutputProfileIs_sRGB = false;
	return rvOk;
}

void CColorProfile::TransformImage( wxImage &img, const cmsHPROFILE inputProfile )
{
	wxASSERT( m_OutputProfile != NULL );
	wxASSERT( inputProfile != NULL );
	
	// transform
	cmsHTRANSFORM hTransform = cmsCreateTransform( inputProfile, TYPE_RGB_8, m_OutputProfile, TYPE_RGB_8, INTENT_PERCEPTUAL, 0 );
	wxASSERT( hTransform != NULL );

	// do the transformation
	cmsUInt32Number imgSize = img.GetWidth() * img.GetHeight();
	cmsDoTransform( hTransform, img.GetData(), img.GetData(), imgSize );

	// close handles
	cmsDeleteTransform( hTransform );
}

void CColorProfile::TransformImage( wxImage &img, const CMemoryChunk &colorProfile )
{
	// we handle images without profile as sRGB, so if there is no input profile
	// and the output profile is sRGB we can return since the transform would do nothing
	if( colorProfile.GetAddress() == NULL && m_OutputProfileIs_sRGB )
		return;

	// input profile
	cmsHPROFILE inputProfile;
	if( colorProfile.GetAddress() != NULL ) {
		// use the given profile
		inputProfile = cmsOpenProfileFromMem( colorProfile.GetAddress(), colorProfile.GetSize() );
	}
	else {
		// no embedded profile, use sRGB
		inputProfile = cmsCreate_sRGBProfile();
	}

	// transform
	TransformImage( img, inputProfile );

	cmsCloseProfile( inputProfile );
}

void CColorProfile::TransformImageFromAdobeRGB( wxImage &img )
{
	TransformImage( img, m_AdobeRGBProfile );
}

cmsHPROFILE CColorProfile::Create_AdobeRGB()
{
    cmsToneCurve* Curve[3];
    cmsHPROFILE hProfile;
    cmsCIExyY D65;
    cmsCIExyYTRIPLE Primaries = {{0.64, 0.33, 1 },
                                 {0.21, 0.71, 1 },
                                 {0.15, 0.06, 1 }};
    
    Curve[0] = Curve[1] = Curve[2] = cmsBuildGamma( NULL, 2.19921875 );

    cmsWhitePointFromTemp(&D65, 6504);
    hProfile = cmsCreateRGBProfile( &D65, &Primaries, Curve );
    cmsFreeToneCurve(Curve[0]);
    
    return hProfile;
}


