/*
	This file is part of vvvP (Virtual Volumes View Photo edition)

	Copyright (C) 2009, the vvvP Development team

	Author: Fulvio Senore

    vvvP is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    vvvP is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with vvvP; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

// For compilers that support precompilation, includes "wx/wx.h".
#include "wx/wxprec.h"

#ifdef __BORLANDC__
#pragma hdrstop
#endif

#ifndef WX_PRECOMP
#include "wx/wx.h"
#endif

#include "image_preview.h"
#include "utils.h"
#include "fs_image.h"
#include "catalog_images_functions.h"

#include <exiv2/preview.hpp>

#include <wx/mstream.h>

CImagePreview::ReturnValue CImagePreview::ExtractImagePreview( const wxString &fileName, int requestedSize, wxImage &previewImage, CMemoryChunk &colorProfile, CMetadata::ColorSpace &exifColorSpace )
{
	try {
		Exiv2::Image::AutoPtr image = Exiv2::ImageFactory::open( CUtils::wx2std(fileName) );
		image->readMetadata();
		Exiv2::PreviewManager pm(*image);
		Exiv2::PreviewPropertiesList list = pm.getPreviewProperties();

		if( list.empty() ) {
			// no previews
			return rtNoPreviews;
		}

		size_t imgIndex;
		if( requestedSize == IMAGE_PREVIEW_LARGEST_SIZE ) {
			imgIndex = list.size() - 1;
		}
		else {
			bool found = false;
			for( size_t k = 0; k < list.size(); k++ ) {
				if( (int) list[k].width_ >= requestedSize || (int) list[k].height_ >= requestedSize ) {
					found = true;
					imgIndex = k;
					break;
				}
			}
			if( !found ) {
				// no previews large enough
				return rtPreviewsTooSmall;
			}
		}

		Exiv2::PreviewImage preview = pm.getPreviewImage( list[imgIndex] );

		// convert the image buffer to a wxImage and resize it
		wxMemoryInputStream mis( preview.pData(), preview.size() );
		{
			wxLogNull ln;
			fsImage img( mis );
			
			if( !img.IsOk() )
				return rtPreviewCorrupted;

			colorProfile.SetData( img.GetICCProfileAddress(), img.GetICCProfileSize() );
			img.DeleteICCProfile();
			previewImage = img;
		}

		// read color space
		CUtils::ImageRotation imageRotation;
		wxDateTime imageDateTime;
		int imageWidth = -1, imageHeight = -1;
		CMetadata::ReadBasicMetadata( preview.pData(), preview.size(), imageRotation, imageDateTime, imageWidth, imageHeight, exifColorSpace );

		return rtOk;
	}
	catch(...) {
		return rtNoPreviews;
	}
}

CImagePreview::ReturnValue CImagePreview::ExtractLargestPreview( const wxString &fileName, wxImage &previewImage, CMemoryChunk &colorProfile, CMetadata::ColorSpace &exifColorSpace )
{
	try {
		Exiv2::Image::AutoPtr image = Exiv2::ImageFactory::open( CUtils::wx2std(fileName) );
		image->readMetadata();
		Exiv2::PreviewManager pm(*image);
		Exiv2::PreviewPropertiesList list = pm.getPreviewProperties();

		if( list.empty() ) {
			// no previews
			return rtNoPreviews;
		}

		size_t imgIndex;
		imgIndex = list.size() - 1;

		Exiv2::PreviewImage preview = pm.getPreviewImage( list[imgIndex] );

		// convert the image buffer to a wxImage and resize it
		previewImage = wxNullImage;
		wxMemoryInputStream mis( preview.pData(), preview.size() );
		{
			wxLogNull ln;
			fsImage img( mis );
			colorProfile.SetData( img.GetICCProfileAddress(), img.GetICCProfileSize() );
			img.DeleteICCProfile();
			previewImage = img;
		}

		if( !previewImage.IsOk() )
			return rtNoPreviews;

		// read color space
		CUtils::ImageRotation imageRotation;
		wxDateTime imageDateTime;
		int imageWidth = -1, imageHeight = -1;
		CMetadata::ReadBasicMetadata( preview.pData(), preview.size(), imageRotation, imageDateTime, imageWidth, imageHeight, exifColorSpace );

		return rtOk;
	}
	catch(...) {
		return rtNoPreviews;
	}
}

CImagePreview::ReturnValue CImagePreview::ExtractLargestPreview( const wxString &fileNameIn, wxString &pathOut )
{
	try {
		Exiv2::Image::AutoPtr image = Exiv2::ImageFactory::open( CUtils::wx2std(fileNameIn) );
		image->readMetadata();
		Exiv2::PreviewManager pm(*image);
		Exiv2::PreviewPropertiesList list = pm.getPreviewProperties();

		if( list.empty() ) {
			// no previews
			return rtNoPreviews;
		}

		size_t imgIndex;
		imgIndex = list.size() - 1;

		Exiv2::PreviewImage preview = pm.getPreviewImage( list[imgIndex] );

		// create the output file name (excluding extension)
		wxFileName fnIn( fileNameIn );
//		wxFileName fnOut( fnIn.GetPath(), fnIn.GetName(), CUtils::std2wx(preview.extension()) );
		wxString fileNameOut = pathOut + wxFileName::GetPathSeparator() + fnIn.GetName();
//		wxString fileNameOut = fnOut.GetFullPath();

		// write image file
		preview.writeFile( CUtils::wx2std( fileNameOut ) );

		return rtOk;
	}
	catch(...) {
		return rtNoPreviews;
	}
}


