/*
	This file is part of vvvP (Virtual Volumes View Photo edition)

	Copyright (C) 2008, the vvvP Development team

	Author: Fulvio Senore

    vvvP is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    vvvP is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with vvvP; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

// For compilers that support precompilation, includes "wx/wx.h".
#include "wx/wxprec.h"

#ifdef __BORLANDC__
#pragma hdrstop
#endif

#ifndef WX_PRECOMP
#include "wx/wx.h"
#endif

#include "metadata.h"
#include "sqlstrings.h"

long CMetadata::nItems = 0;
wxString *CMetadata::metadataDescriptions = NULL;

CMetadata::CMetadata(void)
{
	nItems++;
	
	if( nItems == 1 ) {
		// these are static data that have to be set up only once
		metadataDescriptions = new wxString[GetTagsNumber() + 1];
		SetMetatadaDescriptions();
	}
	
	Clear();
}

CMetadata::~CMetadata(void) {
	if( nItems == 1 ) {
		delete [] metadataDescriptions;
	}

	nItems--;
}


void CMetadata::SetMetatadaDescriptions() {

	metadataDescriptions[1] = _("Camera make");
	metadataDescriptions[2] = _("Camera model");
	metadataDescriptions[3] = _("Image timestamp");
	metadataDescriptions[4] = _("Exposure time");
	metadataDescriptions[5] = _("Aperture");
	metadataDescriptions[6] = _("Exposure bias");
	metadataDescriptions[7] = _("Focal length");
	metadataDescriptions[8] = _("35 mm equivalent");
	metadataDescriptions[9] = _("ISO speed");
	metadataDescriptions[10] = _("Exposure mode");
	metadataDescriptions[11] = _("Metering mode");
	metadataDescriptions[12] = _("Image quality");
	metadataDescriptions[13] = _("White balance");
	metadataDescriptions[14] = _("Max aperture");
	metadataDescriptions[15] = _("Flash");
	metadataDescriptions[16] = _("Software");
}

wxString CMetadata::ReadStringTag( const Exiv2::ExifData& exifData, const wxString &key ) {
	wxString retVal = wxEmptyString;

	Exiv2::ExifKey ek( CUtils::wx2std(key) );
    Exiv2::ExifData::const_iterator md = exifData.findKey(ek);
	if( md != exifData.end() ) {
		retVal = CUtils::std2wx( md->print() );
	}

	return retVal;
}

long CMetadata::ReadLongTag( const Exiv2::ExifData& exifData, const wxString &key )
{
	long retVal = -1;

	Exiv2::ExifKey ek( CUtils::wx2std(key) );
    Exiv2::ExifData::const_iterator md = exifData.findKey(ek);
	if( md != exifData.end() ) {
		retVal = md->toLong();
	}

	return retVal;
}

void CMetadata::AppendMetadataRow( FSDB::Recordset rs, int id, int type, wxString value ) {
	rs.AddNew();
	rs["FILE_ID_MD"] = id;
	rs["METADATA_TYPE"] = type;
	rs["METADATA_VALUE"] = CUtils::DBwx2std( value.Left(50) );
	rs.Update();
}

void CMetadata::ReadBasicMetadata( Exiv2::Image::AutoPtr image, CUtils::ImageRotation &rotation, wxDateTime &imageDateTime, int &width, int &height, ColorSpace &colorSpace )
{
	// do not read metadata in case of error (wrong file type)
	try {
		wxASSERT( image.get() != 0 );
		image->readMetadata();

		// read image width and height
		width = image->pixelWidth();
		height = image->pixelHeight();

		Exiv2::ExifData &exifData = image->exifData();

		if( exifData.empty() ) return;	// no metadata for this file

		// read image orientation
		{
			int exifOrientation = 1;
			int tmpOrientation = ReadLongTag( exifData, wxT("Exif.Image.Orientation") );
			if( tmpOrientation != -1 )
				exifOrientation = tmpOrientation;
			switch( exifOrientation ) {
				case 3:
					rotation = CUtils::irUpsideDown;
					break;
				case 6:
					rotation = CUtils::irRotateRight;
					break;
				case 8:
					rotation = CUtils::irRotateLeft;
					break;
			}
		}

		// image timestamp
		wxString s = ReadStringTag( exifData, wxT("Exif.Photo.DateTimeOriginal") );
		if( !s.empty() ) {
			imageDateTime = DecodeTimestamp( CUtils::wx2std(s) );
		}

		// color space
		long ltmp = ReadLongTag( exifData, wxT("Exif.Photo.ColorSpace") );
		switch( ltmp ) {
			case 1:
				colorSpace = cs_sRGB;
				break;
			case 2:
				// non standar but used sometime
				colorSpace = csAdobeRGB;
				break;
			case 65535:
				// uncalibrated, see interop index
				s = ReadStringTag( exifData, wxT("Exif.Iop.InteroperabilityIndex") );
				if( s == wxT("R03") ) {
					colorSpace = csAdobeRGB;
				}
				break;
		}

	}
	catch(...) {
		return;
	}
}

void CMetadata::ReadBasicMetadata( wxString fileName, CUtils::ImageRotation &rotation, wxDateTime &imageDateTime, int &width, int &height, ColorSpace &colorSpace )
{
	rotation = CUtils::irNoRotation;
	imageDateTime = wxInvalidDateTime;
	width = -1;
	height = -1;
	colorSpace = csUnknown;

	// do not read metadata in case of error (wrong file type)
	try {
		Exiv2::Image::AutoPtr image = Exiv2::ImageFactory::open( CUtils::wx2std(fileName) );
		ReadBasicMetadata( image, rotation, imageDateTime, width, height, colorSpace );
	}
	catch(...) {
		return;
	}
}

void CMetadata::ReadBasicMetadata( const unsigned char *bufferData, long bufferSize, CUtils::ImageRotation &rotation, wxDateTime &imageDateTime, int &width, int &height, ColorSpace &colorSpace )
{
	rotation = CUtils::irNoRotation;
	imageDateTime = wxInvalidDateTime;
	width = -1;
	height = -1;
	colorSpace = csUnknown;

	// do not read metadata in case of error (wrong file type)
	try {
		Exiv2::Image::AutoPtr image = Exiv2::ImageFactory::open( bufferData, bufferSize );
		ReadBasicMetadata( image, rotation, imageDateTime, width, height, colorSpace );
	}
	catch(...) {
		return;
	}
}

void CMetadata::ReadFileTags( wxString fileName, int fileID, FSDB::Database db ) {

	try {
		Exiv2::Image::AutoPtr image = Exiv2::ImageFactory::open( CUtils::wx2std(fileName) );
		wxASSERT( image.get() != 0 );
		image->readMetadata();
		Exiv2::ExifData &exifData = image->exifData();
		if( exifData.empty() ) return;	// no metadata for this file

		FSDB_CREATE_RS_READWRITE_FWDONLY( db, rs );
		std::string sql = CSqlStrings::SelectMetadataEmpty();
		rs.Open( sql, "METADATA", "METADATA_ID" );

		wxString s;

		// camera make
		s = ReadStringTag( exifData, wxT("Exif.Image.Make") );
		if( !s.empty() ) AppendMetadataRow( rs, fileID, mtCameraMake, s );

		// camera model
		s = ReadStringTag( exifData, wxT("Exif.Image.Model") );
		if( !s.empty() ) AppendMetadataRow( rs, fileID, mtCameraModel, s );

		// image timestamp
		s = ReadStringTag( exifData, wxT("Exif.Photo.DateTimeOriginal") );
		if( !s.empty() ) {
			AppendMetadataRow( rs, fileID, mtImageTimestamp, s );
		}

		// exposure time
		s = ReadStringTag( exifData, wxT("Exif.Photo.ExposureTime") );
		if( s.empty() ) s = ReadStringTag( exifData, wxT("Exif.Photo.ShutterSpeedValue") );
		if( !s.empty() ) {
			// remove the trailing " s"
			int l = (int) s.length();
			if( (s[l - 2] == ' ') && ((s[l - 1] == 's') || (s[l - 1] == 'S')) ) {
				s = s.substr( 0, l - 2 );
			}
			AppendMetadataRow( rs, fileID, mtExposureTime, s );
		}

		// aperture
		s = ReadStringTag( exifData, wxT("Exif.Photo.FNumber") );
		if( s.empty() ) s = ReadStringTag( exifData, wxT("Exif.Photo.ApertureValue") );
		if( !s.empty() ) {
			// remove the leading "F"
			int l = (int) s.length();
			if( (s[0] == 'F') || (s[0] == 'f') ) {
				s = s.substr( 1, l - 1 );
			}
			AppendMetadataRow( rs, fileID, mtAperture, s );
		}

		// exposure bias
		s = ReadStringTag( exifData, wxT("Exif.Photo.ExposureBiasValue") );
		if( !s.empty() ) AppendMetadataRow( rs, fileID, mtExposureBias, s );

		// focal lenght
		s = ReadStringTag( exifData, wxT("Exif.Photo.FocalLength") );
		if( !s.empty() ) {
			AppendMetadataRow( rs, fileID, mtFocalLength, s );
			s = ReadStringTag( exifData, wxT("Exif.Photo.FocalLengthIn35mmFilm") );
			if( !s.empty() ) AppendMetadataRow( rs, fileID, mtFocalLenght35mm, s );
		}

		// ISO speed
		// from ISOSpeedRatings or the Makernote
		s = ReadStringTag( exifData, wxT("Exif.Photo.ISOSpeedRatings") );
		if( s.empty() ) s = ReadStringTag( exifData, wxT("Exif.CanonSi.ISOSpeed") );
		if( s.empty() ) s = ReadStringTag( exifData, wxT("Exif.Nikon1.ISOSpeed") );
		if( s.empty() ) s = ReadStringTag( exifData, wxT("Exif.Nikon2.ISOSpeed") );
		if( s.empty() ) s = ReadStringTag( exifData, wxT("Exif.Nikon3.ISOSpeed") );
		if( s.empty() ) s = ReadStringTag( exifData, wxT("Exif.MinoltaCsNew.ISOSpeed") );
		if( s.empty() ) s = ReadStringTag( exifData, wxT("Exif.MinoltaCsOld.ISOSpeed") );
		if( s.empty() ) s = ReadStringTag( exifData, wxT("Exif.MinoltaCs5D.ISOSpeed") );
		if( s.empty() ) s = ReadStringTag( exifData, wxT("Exif.MinoltaCs7D.ISOSpeed") );
		if( !s.empty() ) AppendMetadataRow( rs, fileID, mtISOSpeed, s );

		// Exposure mode
		s = ReadStringTag( exifData, wxT("Exif.Photo.ExposureProgram") );
		if( s.empty() ) s = ReadStringTag( exifData, wxT("Exif.CanonCs.ExposureProgram") );
		if( s.empty() ) s = ReadStringTag( exifData, wxT("Exif.MinoltaCs7D.ExposureMode") );
		if( s.empty() ) s = ReadStringTag( exifData, wxT("Exif.MinoltaCs5D.ExposureMode") );
		if( s.empty() ) s = ReadStringTag( exifData, wxT("Exif.MinoltaCsNew.ExposureMode") );
		if( s.empty() ) s = ReadStringTag( exifData, wxT("Exif.MinoltaCsOld.ExposureMode") );
		if( !s.empty() ) AppendMetadataRow( rs, fileID, mtExposureMode, s );

		// metering mode
		s = ReadStringTag( exifData, wxT("Exif.Photo.MeteringMode") );
		if( !s.empty() ) AppendMetadataRow( rs, fileID, mtMeteringMode, s );

		// image quality
		s = ReadStringTag( exifData, wxT("Exif.CanonCs.Quality") );
		if( s.empty() ) s = ReadStringTag( exifData, wxT("Exif.Fujifilm.Quality") );
		if( s.empty() ) s = ReadStringTag( exifData, wxT("Exif.Sigma.Quality") );
		if( s.empty() ) s = ReadStringTag( exifData, wxT("Exif.Nikon1.Quality") );
		if( s.empty() ) s = ReadStringTag( exifData, wxT("Exif.Nikon2.Quality") );
		if( s.empty() ) s = ReadStringTag( exifData, wxT("Exif.Nikon3.Quality") );
		if( s.empty() ) s = ReadStringTag( exifData, wxT("Exif.Olympus.Quality") );
		if( s.empty() ) s = ReadStringTag( exifData, wxT("Exif.Panasonic.Quality") );
		if( s.empty() ) s = ReadStringTag( exifData, wxT("Exif.Minolta.Quality") );
		if( s.empty() ) s = ReadStringTag( exifData, wxT("Exif.MinoltaCsNew.Quality") );
		if( s.empty() ) s = ReadStringTag( exifData, wxT("Exif.MinoltaCsOld.Quality") );
		if( s.empty() ) s = ReadStringTag( exifData, wxT("Exif.MinoltaCs5D.Quality") );
		if( s.empty() ) s = ReadStringTag( exifData, wxT("Exif.MinoltaCs7D.Quality") );
		if( !s.empty() ) AppendMetadataRow( rs, fileID, mtImageQuality, s );

		// white balance
		s = ReadStringTag( exifData, wxT("Exif.CanonSi.WhiteBalance") );
		if( s.empty() ) s = ReadStringTag( exifData, wxT("Exif.Fujifilm.WhiteBalance") );
		if( s.empty() ) s = ReadStringTag( exifData, wxT("Exif.Sigma.WhiteBalance") );
		if( s.empty() ) s = ReadStringTag( exifData, wxT("Exif.Nikon1.WhiteBalance") );
		if( s.empty() ) s = ReadStringTag( exifData, wxT("Exif.Nikon2.WhiteBalance") );
		if( s.empty() ) s = ReadStringTag( exifData, wxT("Exif.Nikon3.WhiteBalance") );
		if( s.empty() ) s = ReadStringTag( exifData, wxT("Exif.Olympus.WhiteBalance") );
		if( s.empty() ) s = ReadStringTag( exifData, wxT("Exif.Panasonic.WhiteBalance") );
		if( s.empty() ) s = ReadStringTag( exifData, wxT("Exif.MinoltaCs5D.WhiteBalance") );
		if( s.empty() ) s = ReadStringTag( exifData, wxT("Exif.MinoltaCs7D.WhiteBalance") );
		if( s.empty() ) s = ReadStringTag( exifData, wxT("Exif.MinoltaCsNew.WhiteBalance") );
		if( s.empty() ) s = ReadStringTag( exifData, wxT("Exif.MinoltaCsOld.WhiteBalance") );
		if( !s.empty() ) AppendMetadataRow( rs, fileID, mtWhiteBalance, s );

		// max aperture
		s = ReadStringTag( exifData, wxT("Exif.Photo.MaxApertureValue") );
		if( !s.empty() ) {
			// remove the leading "F"
			int l = (int) s.length();
			if( (s[0] == 'F') || (s[0] == 'f') ) {
				s = s.substr( 1, l - 1 );
			}
			AppendMetadataRow( rs, fileID, mtMaxAperture, s );
		}

		// flash
		s = ReadStringTag( exifData, wxT("Exif.Photo.Flash") );
		if( !s.empty() ) AppendMetadataRow( rs, fileID, mtFlash, s );

		// software
		s = ReadStringTag( exifData, wxT("Exif.Image.Software") );
		if( !s.empty() ) AppendMetadataRow( rs, fileID, mtSoftware, s );

		rs.Close();
	}
	catch(...) {
		return;
	}
}

wxDateTime CMetadata::DecodeTimestamp( std::string exifValue ) {

	if( exifValue == "" || exifValue[0] == ' ' || exifValue[0] == '0' ) {
		// data non set, return an empty object
		return wxDateTime();
	}

	wxString s = CUtils::std2wx( exifValue );
	long y, mo, d, h, mi, sec;
	s.Left(4).ToLong( &y );
	s.Mid(5, 2).ToLong( &mo );
	s.Mid(8, 2).ToLong( &d );
	s.Mid(11, 2).ToLong( &h );
	s.Mid(14, 2).ToLong( &mi );
	s.Mid(17, 2).ToLong( &sec );

	return wxDateTime( d, (wxDateTime::Month) (mo - 1), y, h, mi, sec );

}

void CMetadata::LoadSingleTagFromDB( FSDB::Recordset rs ) {

	MetadataType mt = (MetadataType) (int) rs["METADATA_TYPE"];
	switch( mt ) {
		case mtCameraMake:
			m_CameraMake = CUtils::DBstd2wx( rs["METADATA_VALUE"] );
			break;
		case mtCameraModel:
			m_CameraModel = CUtils::DBstd2wx( rs["METADATA_VALUE"] );
			break;
		case mtSoftware:
			m_Software = CUtils::DBstd2wx( rs["METADATA_VALUE"] );
			break;
		case mtExposureTime:
			m_ExposureTime = CUtils::DBstd2wx( rs["METADATA_VALUE"] );
			break;
		case mtAperture:
			m_Aperture = CUtils::DBstd2wx( rs["METADATA_VALUE"] );
			break;
		case mtMaxAperture:
			m_MaxAperture = CUtils::DBstd2wx( rs["METADATA_VALUE"] );
			break;
		case mtExposureBias:
			m_ExposureBias = CUtils::DBstd2wx( rs["METADATA_VALUE"] );
			break;
		case mtFlash:
			m_Flash = CUtils::DBstd2wx( rs["METADATA_VALUE"] );
			break;
		case mtFocalLength:
			m_FocalLength = CUtils::DBstd2wx( rs["METADATA_VALUE"] );
			break;
		case mtFocalLenght35mm:
			m_FocalLength35mm = CUtils::DBstd2wx( rs["METADATA_VALUE"] );
			break;
		case mtISOSpeed:
			m_ISOSpeed = CUtils::DBstd2wx( rs["METADATA_VALUE"] );
			break;
		case mtExposureMode:
			m_ExposureMode = CUtils::DBstd2wx( rs["METADATA_VALUE"] );
			break;
		case mtMeteringMode:
			m_MeteringMode = CUtils::DBstd2wx( rs["METADATA_VALUE"] );
			break;
		case mtImageQuality:
			m_ImageQuality = CUtils::DBstd2wx( rs["METADATA_VALUE"] );
			break;
		case mtWhiteBalance:
			m_WhiteBalance = CUtils::DBstd2wx( rs["METADATA_VALUE"] );
			break;
		case mtImageTimestamp:
			m_ImageTimestamp = DecodeTimestamp( rs["METADATA_VALUE"] );
			break;
	}
}

void CMetadata::LoadFromDB( int fileID, FSDB::Database db ) {

	Clear();

	FSDB_CREATE_RS_READONLY_FWDONLY( db, rs );
	std::string sql = CSqlStrings::SelectMetadataFileId( fileID );
	rs.Open( sql );

	while( !rs.IsEOF() ) {
		LoadSingleTagFromDB( rs );
		rs.MoveNext();
	}

	rs.Close();
}

void CMetadata::Clear() {
	m_CameraMake = wxEmptyString;
	m_CameraModel = wxEmptyString;
	m_Software = wxEmptyString;
	m_ExposureTime = wxEmptyString;
	m_Aperture = wxEmptyString;
	m_MaxAperture = wxEmptyString;
	m_ExposureBias = wxEmptyString;
	m_Flash = wxEmptyString;
	m_FocalLength = wxEmptyString;
	m_FocalLength35mm = wxEmptyString;
	m_ISOSpeed = wxEmptyString;
	m_ExposureMode = wxEmptyString;
	m_MeteringMode = wxEmptyString;
	m_ImageQuality = wxEmptyString;
	m_WhiteBalance = wxEmptyString;
	m_ImageTimestamp = wxInvalidDateTime;
}

void CMetadata::DeleteFileTags( int fileID, FSDB::Database db ) {
	std::string sql = CSqlStrings::DeleteFileMetadata( fileID );
	db->ExecuteNoReturn( sql );
}

