/*
	This file is part of vvvP (Virtual Volumes View Photo edition)

	Copyright (C) 2008, the vvvP Development team

	Author: Fulvio Senore

    vvvP is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    vvvP is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with vvvP; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#ifndef __VVVP_THUMBNAILCTRL_H__
#define __VVVP_THUMBNAILCTRL_H__

#include "wx/bitmap.h"
#include "wx/scrolwin.h"

#include "thumbnailctrl.h"
#include "utils.h"


// class used by vvvP
class vvvPImageThumbnailItem: public wxImageThumbnailItem {

    DECLARE_CLASS(vvvPImageThumbnailItem)

public:

	// receives a reference to and object containing data about this item
	vvvPImageThumbnailItem( CFileData* fileData );

    /// Load the thumbnail. It will not create the thumbnail bitmap to gain speed: see Draw()
    virtual bool Load(wxThumbnailCtrl* ctrl, bool forceLoad) ;

    /// Draw the background. Overridden because we need to show data of this specific class
    virtual bool DrawBackground(wxDC& dc, wxThumbnailCtrl* ctrl, const wxRect& rect, const wxRect& imageRect, int style, int index) ;

	/// Draw the item. Overridden because it creates the cached bitmap the first time that the object is drawn.
	/// this adds spees because  the original version created all the bitmaps at the same time, slowing down things
    virtual bool Draw(wxDC& dc, wxThumbnailCtrl* ctrl, const wxRect& rect, int style) ;

	// return a pointer to the CFileData object
	CFileData* GetFileData() { return m_fileData; }

protected:

	/// create a the bitmap that will be drawn to the screen
	void CreateCachedBitmap( wxThumbnailCtrl* ctrl );

	CFileData* m_fileData;

	// true if we need to create the thumbnail bitmap. It happens at object creation and when the thumbnail size is changed.
	bool m_isThumbnailDirty;
};



class vvvPThumbnailCtrl : public wxThumbnailCtrl {

    DECLARE_CLASS(vvvPThumbnailCtrl)
    DECLARE_EVENT_TABLE()

public:

    vvvPThumbnailCtrl( wxWindow* parent, wxWindowID id = -1, const wxPoint& pos = wxDefaultPosition, const wxSize& size = wxDefaultSize,
        long style = wxTH_TEXT_LABEL|wxTH_IMAGE_LABEL|wxTH_EXTENSION_OR_DATETIME_LABEL );

	// rotate the currently selected images by the given amount
	// also update the database
	void RotateSelectedThumbnails( CUtils::ImageRotation rotation );

	/// Return true if one or more items are selected
	bool IsAnythingSelected();

	/// return a vector contining the 0-based indexes of the selected items 
	/// or an empty vector if nothing is selected
	std::vector<int> GetSelectedItemsIndexes();

	/// return a vector containing the pointers to the CFileData of all the selected items
	/// or an empty vector if nothing is selected
	std::vector<CFileData*> GetSelectedItemsCFileData();

	/// show the "image information" dialog if a single image is selected,
	/// or the "object description" dialog if multiple images are selected
	void ShowInfoDialog();

	/// Fill the control with the data contained in the container
	void ShowContainerData( CFileDataArray &fileDataArray );

	/// Prepare the control for loading data (clears everything)
	void PrepareForLoading();

	/// Append new images to the control
	/// if fileDataArray contains more items than the control appends the new items
	/// the code is based only on the number of elements
	/// the caller MUST always pass the same fileDataArray afted adding some elements
	void LoadNewData( CFileDataArray &fileDataArray );

	/// return the total number of files shown in the control
	int GetNumFilesTot() { return numFilesTot; }
	/// return the total size of the files shown in the control
	wxULongLong GetSizeFilesTot() { return sizeFilesTot; }

	/// return the index of the first visible item, or -1 if unable to find it
	int GetFirstVisibleItemIndex();

    /// The size of the image part
    void SetThumbnailImageSize(const wxSize& sz);

    /// Append a single item
    virtual int Append(wxThumbnailItem* item);

protected:

	/// event handler for a context menu request
	void OnContextMenu( wxContextMenuEvent& event );

    void OnSetFocus( wxFocusEvent& event );

	// rotate a single image by the given amount
	// nThumbnail is the index of the image
	void RotateSingleThumbnail( CUtils::ImageRotation rotation, int nThumbnail );

	// total number of files shown in the control
	int numFilesTot;
	// total size of the files shown in the control
	wxULongLong sizeFilesTot;
};










#endif

